"""
Centralized type handling utilities for extract methods.
"""

import logging
from typing import get_args, get_origin

from src.baml_client.types import ExtractedData
from pydantic import BaseModel

PRIMITIVE_TYPES = (int, float, str, bool, type(None))
COLLECTION_TYPES = (list,)
logger = logging.getLogger(__name__)


def is_primitive_type(python_type: type) -> bool:
    """Check if a type is a supported primitive."""
    return python_type in PRIMITIVE_TYPES


def is_collection_type(python_type: type) -> bool:
    """Check if a type is a supported collection."""
    return python_type in COLLECTION_TYPES


def is_basemodel_type(python_type: type) -> bool:
    """Check if a type is a BaseModel subclass."""
    try:
        return isinstance(python_type, type) and issubclass(python_type, BaseModel)
    except TypeError:
        return False


def convert_extracted_data(
    schema: ExtractedData, output: ExtractedData
) -> ExtractedData:
    """
    Convert extracted data to appropriate Python type.

    Args:
        schema: The target schema type
        extracted_data: Raw data from BAML extraction

    Returns:
        Converted data matching schema type
    """
    logger.debug(f"Converting extracted data for schema: {schema}, output: {output}")

    # Handle generic types like list[BaseModel]
    origin = get_origin(schema)
    logger.debug(f"Origin: {origin}")
    if origin is list:
        args = get_args(schema)
        logger.debug(f"List args: {args}, ")
        if args and is_basemodel_type(args[0]):
            # Handle list[BaseModel]
            model_class = args[0]
            extracted_data = output.model_dump(by_alias=False).get("schema", [])
            return [model_class.model_validate(item, by_alias=False, by_name=True) for item in extracted_data]
        else:
            return list(output) if output else []

    if is_primitive_type(schema):
        if schema is str:
            return str(output)
        elif schema is int:
            return int(output)
        elif schema is float:
            return float(output)
        elif schema is bool:
            return bool(output)
        elif schema is type(None):
            return None
    elif is_collection_type(schema):
        if schema is list:
            return list(output) if output else []
    elif is_basemodel_type(schema):
        # NOTE: Sometimes model does not annotate as list[Type], but just Type
        # And it returns a list at the end :).
        if isinstance(output, list):
            res = []
            for item in output:
                extracted_data = item.model_dump(by_alias=False).get("schema", {})
                res.append(schema.model_validate(extracted_data, by_alias=False, by_name=True))
            return res
        else:
            extracted_data = output.model_dump(by_alias=False).get("schema", {})
            logger.debug(
                f"Checking type of extracted data {type(extracted_data)} {extracted_data} {schema=}"
            )
            return schema.model_validate(extracted_data, by_alias=False, by_name=True)

    return output



# ----------------------------------------------------------------------------
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
# ----------------------------------------------------------------------------

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code using: baml-cli generate
# baml-cli is available with the baml package.

_file_map = {

    "clients.baml": "// Learn more about clients at https://docs.boundaryml.com/docs/snippets/clients/overview\n\n// This is a stub client definition.\n// Actual provider clients (with real implementations) are selected and\n// instantiated at runtime based on the config/env\nclient<llm> Base {\n  provider \"openai-generic\"\n  options {\n    base_url \"\"\n  }\n}\n\nclient<llm> GPT_QQ {\n  provider openai\n  retry_policy Exponential\n  options {\n    model \"gpt-4o-mini\"\n    base_url \"https://api.key77qiqi.com/v1\"\n    api_key env.OPENAI_QQ_API_KEY\n    temperature 0.0\n  }\n}\n\nclient<llm> GEMINI_QQ {\n  provider google-ai\n  retry_policy Exponential\n  options {\n    model \"gemini-2.5-flash\"\n    base_url \"https://api.key77qiqi.com/v1\"\n    api_key env.OPENAI_QQ_API_KEY\n    temperature 0.0\n  }\n}\n\nclient<llm> GPTOfficial {\n  provider openai\n  // retry_policy Exponential\n  options {\n    model \"gpt-4o-mini\"\n    api_key env.OPENAI_API_KEY\n    temperature 0.0\n  }\n}\n\nclient<llm> Gemini {\n  provider google-ai\n  retry_policy Exponential\n  options {\n    model \"gemini-2.5-pro\"\n    api_key env.GOOGLEAI_API_KEY\n  }\n}\n\nclient<llm> Claude {\n  provider anthropic\n  retry_policy Exponential\n  options {\n    model \"claude-sonnet-4-20250514\"\n    api_key env.ANTHROPIC_API_KEY\n  }\n}\n\nclient<llm> Qwen72b {\n  provider \"openai-generic\"\n  options {\n    base_url \"http://localhost:8000/v1\"\n    model \"Qwen/Qwen2.5-VL-72B-Instruct-AWQ\"\n  }\n}\n\nclient<llm> Qwen32b {\n  provider \"openai-generic\"\n  options {\n    base_url \"http://localhost:8000/v1\"\n    model \"Qwen/Qwen2.5-VL-32B-Instruct-AWQ\"\n  }\n}\n\nclient<llm> Qwen7b {\n  provider \"openai-generic\"\n  options {\n    base_url \"http://localhost:8000/v1\"\n    model \"Qwen/Qwen2.5-VL-7B-Instruct-AWQ\"\n  }\n}\n\nclient<llm> Qwen3b {\n  provider \"openai-generic\"\n  options {\n    base_url \"http://localhost:8000/v1\"\n    model \"Qwen/Qwen2.5-VL-3B-Instruct-AWQ\"\n  }\n}\n\nclient<llm> OpenRouter {\n  provider \"openai-generic\"\n  options {\n    base_url \"https://openrouter.ai/api/v1\"\n    model \"openai/gpt-oss-120b\"\n    api_key env.OPENROUTER_API_KEY\n  }\n}\n\nclient<llm> ZrokLocalModel {\n  provider \"openai-generic\"\n  retry_policy Exponential\n  options {\n    base_url \"https://llms.zrok.06946099.xyz/v1\"\n    model \"Qwen/Qwen3-VL-30B-A3B-Thinking\"\n    temperature 0.0\n  }\n}\n\nclient<llm> NgrokShowUI {\n  provider \"openai-generic\"\n  // retry_policy Exponential\n  options {\n    base_url \"https://viperish-fawn-superhumanly.ngrok-free.dev/v1\"\n    model \"showlab/ShowUI-2B\"\n    // temperature 0.0\n  }\n}\n\nclient<llm> NgrokLocalModel {\n  provider \"openai-generic\"\n  retry_policy Exponential\n  options {\n    base_url \"https://viperish-fawn-superhumanly.ngrok-free.dev/v1\"\n    model \"Qwen/Qwen3-VL-30B-A3B-Thinking\"\n    temperature 0.0\n    request_timeout_ms 8000\n    connect_timeout_ms 2000\n  }\n}\n\nclient<llm> FallbackLocalModel {\n  provider fallback\n  options {\n    strategy [\n      GPT_QQ\n    ]\n  }\n}\n\nclient<llm> ZrokUIVenus {\n  provider \"openai-generic\"\n  // retry_policy Exponential\n  options {\n    base_url \"https://uivenus.zrok.06946099.xyz/v1\"\n    api_key \"token-abc123\"\n    model \"inclusionAI/UI-Venus-Ground-7B\"\n    temperature 0.0\n    http {\n      request_timeout_ms 8000\n      connect_timeout_ms 2000\n    }\n    header {\n      \"secret\" \"sjtu-course-cs3604-2025\"\n    }\n  }\n}\n\nclient<llm> ZrokUIVenus2 {\n  provider \"openai-generic\"\n  retry_policy Exponential\n  options {\n    base_url \"http://202.120.37.109:23497/v1\"\n    api_key \"token-abc123\"\n    model \"inclusionAI/UI-Venus-Ground-7B\"\n    temperature 0.0\n    http {\n      request_timeout_ms 8000\n      connect_timeout_ms 2000\n    }\n    header {\n      \"secret\" \"sjtu-course-cs3604-2025\"\n    }\n  }\n}\n\nclient<llm> NgrokUIVenus {\n  provider \"openai-generic\"\n  // retry_policy Exponential\n  options {\n    base_url \"https://8dcd0b9d943d.ngrok-free.app/v1\"\n    api_key \"token-abc123\"\n    http {\n      request_timeout_ms 8000\n      connect_timeout_ms 2000\n    }\n    model \"inclusionAI/UI-Venus-Ground-7B\"\n    temperature 0.0\n  }\n}\n\nclient<llm> NgrokUIVenus2 {\n  provider \"openai-generic\"\n  // retry_policy Exponential\n  options {\n    base_url \"https://viperish-fawn-superhumanly.ngrok-free.dev/v1\"\n    api_key \"token-abc123\"\n    http {\n      request_timeout_ms 8000\n      connect_timeout_ms 2000\n    }\n    model \"inclusionAI/UI-Venus-Ground-7B\"\n    temperature 0.0\n  }\n}\n\nclient<llm> FallbackUIVenus1 {\n  provider fallback\n  options {\n    strategy [\n      ZrokUIVenus\n      ZrokUIVenus2\n      NgrokUIVenus\n      NgrokUIVenus2\n    ]\n  }\n}\n\nclient<llm> FallbackUIVenus2 {\n  provider fallback\n  options {\n    strategy [\n      ZrokUIVenus2\n      ZrokUIVenus\n      NgrokUIVenus2\n      NgrokUIVenus\n    ]\n  }\n}\n\nclient<llm> UIVenusRoundRobin {\n  provider round-robin\n  options {\n    strategy [\n      FallbackUIVenus2\n      // FallbackUIVenus1\n    ]\n  }\n}\n\nretry_policy Exponential {\n  max_retries 2\n  strategy {\n    type exponential_backoff\n    delay_ms 300\n    multiplier 1.5\n    max_delay_ms 20000\n  }\n}\n",
    "executor/action.baml": "template_string InternVL_Prompt(screenshot: image, task: string) #\"\n  {{_.role(\"system\")}}\n  You are InternVL, a GUI agent.\n  You are given a task and a screenshot of the screen. You need to perform a series of pyautogui actions to complete the task. \n\n  {{_.role(\"user\")}}\n  {{ task }}\n  {{ screenshot }}\n\"#\n\ntemplate_string UITARS_Prompt(screenshot: image, task: string) #\"\n  {{_.role(\"user\")}}\n  You are a GUI agent. You are given a task and your action history, with screenshots. You need to perform the next action to complete the task.\n\n  ## Output Format\n  ```\n  Thought: ...\n  Action: ...\n  ```\n\n  ## Action Space\n\n  click(point='<point>x1 y1</point>')\n  left_double(point='<point>x1 y1</point>')\n  right_single(point='<point>x1 y1</point>')\n  drag(start_point='<point>x1 y1</point>', end_point='<point>x2 y2</point>')\n  hotkey(key='ctrl c') # Split keys with a space and use lowercase. Also, do not use more than 3 keys in one hotkey action.\n  type(content='xxx') # Use escape characters \\\\', \\\\\\\", and \\\\n in content part to ensure we can parse the content in normal python string format. If you want to submit your input, use \\\\n at the end of content. \n  scroll(point='<point>x1 y1</point>', direction='down or up or right or left') # Show more information on the `direction` side.\n  wait() #Sleep for 5s and take a screenshot to check for any changes.\n  finished(content='xxx') # Use escape characters \\\\', \\\\\", and \\\\n in content part to ensure we can parse the content in normal python string format.\n\n\n  ## Note\n  - Use English in `Thought` part.\n  - Write a small plan and finally summarize your next action (with its target element) in one sentence in `Thought` part.\n\n  ## User Instruction\n  {{ task }}\n  {{ screenshot }}\n\"#\n\ntemplate_string ShowUI_Prompt(screenshot: image, task: string) #\"\n  {{ _.role(\"system\") }}\n  You are an assistant trained to navigate the web screen. \n  Given a task instruction, a screen observation, and an action history sequence, \n  output the next action and wait for the next observation. \n  Here is the action space:\n  1. `CLICK`: Click on an element, value is not applicable and the position [x,y] is required. \n  2. `INPUT`: Type a string into an element, value is a string to type and the position [x,y] is required. \n  3. `SELECT`: Select a value for an element, value is not applicable and the position [x,y] is required. \n  4. `HOVER`: Hover on an element, value is not applicable and the position [x,y] is required.\n  5. `ANSWER`: Answer the question, value is the answer and the position is not applicable.\n  6. `ENTER`: Enter operation, value and position are not applicable.\n  7. `SCROLL`: Scroll the screen, value is the direction to scroll and the position is not applicable.\n  8. `SELECT_TEXT`: Select some text content, value is not applicable and position [[x1,y1], [x2,y2]] is the start and end position of the select operation.\n  9. `COPY`: Copy the text, value is the text to copy and the position is not applicable.\n\n  Format the action as a dictionary with the following keys:\n  {'action': 'ACTION_TYPE', 'value': 'element', 'position': [x,y]}\n\n  If value or position is not applicable, set it as `None`.\n  Position might be [[x1,y1], [x2,y2]] if the action requires a start and end position.\n  Position represents the relative coordinates on the screenshot and should be scaled to a range of 0-1.\n\n  {{_.role(\"user\")}}\n\n  Task: {{ task }}\n  Screenshot: {{ screenshot }}\n\"#\n\ntemplate_string Prompt(screenshot: image, task: string) #\"\n  {{_.role(\"user\")}}\n  You are a UI/UX expert skilled at interacting with websites.\n\n  Given a task and current page screenshot, your goal is to propose one or more of the following actions in Python format, calling the functions directly as shown below (within a Python code block):\n\n  ```python\n  click(target_description=\"...\") # Click on the element matching the description\n  type(target_description=\"...\", content=\"...\") # Focus on the element matching the description and type the specified content\n  drag(source_description=\"...\", target_description=\"...\") # Drag from the source element to the target element, both specified by descriptions\n  scroll(target_description=\"...\", direction=\"...\") # Focus on the described element (or None to scroll the page) and scroll in the given direction (\"up\", \"down\", \"left\", or \"right\")\n  wait(duration=...) # Wait for the specified duration in milliseconds\n  finished() # No action required\n  ```\n\n  Do not provide any explanations in your output; only return the Python code block with the required actions.\n  Target descriptions should be descriptive (i.e. which element, how to find it, what it looks like, ...) for better locating.\n  Target description should comply descriptions in the given task description.\n  When there are similar elements, provide distinguishing features in the description.\n  When selecting from dropdown, don't use `type(...)`.\n\n  Task: {{ task }}\n  Screenshot: {{ screenshot }}\n\"#\n\nfunction ProposeActions(screenshot: image, task: string) -> string {\n  client \"Base\"\n  prompt #\"\n    {% if ctx.client.name == \"InternVL\" %}\n      {{ InternVL_Prompt(screenshot, task) }}\n    {% elif ctx.client.name == \"UITARS\" %}\n      {{ UITARS_Prompt(screenshot, task) }}\n    {% else %}\n      {{ Prompt(screenshot, task) }}\n    {% endif %}\n  \"#\n}\n\nfunction LocateUIElement(screenshot: image, description: string) -> string {\n  client \"Base\"\n  prompt #\"\n    {{_.role(\"user\")}}\n    Outline the position corresponding to the instruction and target description: {{description}}.\n    The output should be only [x1,y1,x2,y2].\n    Screenshot: {{ screenshot }}\n  \"#\n}\n\ntest TestProposeActions {\n  functions [ProposeActions]\n  args {\n    screenshot { url \"https://github.com/user-attachments/assets/624097ea-06f2-4c8f-83f6-b6b9ee439c0c\" }\n    task \"Search the weather for the New York city.\"\n  }\n}\n\ntest TestLocateUIElement {\n  functions [LocateUIElement]\n  args {\n    screenshot { url \"https://github.com/user-attachments/assets/624097ea-06f2-4c8f-83f6-b6b9ee439c0c\" }\n    description \"Search the weather for the New York city.\"\n  }\n}",
    "executor/assertion.baml": "class Feedback {\n    response string\n    reason string\n}\n\nclass History {\n    page_id string\n    layout string?\n    description string?\n    prev_action string?\n}\n\ntemplate_string Specification() #\"\n    # API Specification\n\n    ### Session API\n    - `history -> list[State]`: Chronological sequence of all captured states in the current test session.\n\n    ### State API\n    - `page_id -> str`: Canonical identifier for logical page/state identity.\n    - `title -> str`: Browser tab's visible title.\n    - `url -> str`: Current browser URL.\n    - `extract(instruction: str, schema: Type[BaseModel]) -> Type[BaseModel]`: Extract structured data from the state. `schema` should be a class inherited from `BaseModel`, NOT `BaseModel` itself.\n\n    Important requirements for `schema`:\n      - `schema` must be a subclass of `BaseModel` (a class that inherits from `BaseModel` and defines fields). Do NOT pass `BaseModel` itself.\n      - The subclass should declare the expected fields and their types. Example:\n        ```python\n        class User(BaseModel):\n            id: int\n            name: str\n        ```\n      - The extractor will populate and return an instance of the provided subclass (or data matching its schema). Provide only class objects, not instances or the `BaseModel` base class.\n      - If you provide a schema that is not a subclass of `BaseModel` or omits required fields, extraction will fail. Ensure the schema accurately reflects the structure you expect to extract from the state.\n      - Note that `list` or `dict` or `list[dict]` are not supported types for attributes, please specify inner types properly such as `list[str]`, ...\n      - When want to get a list of items, please define a wrapper model. For example, to extract a list of `User` objects, define:\n        ```python\n        class User(BaseModel):\n            id: int\n            name: str\n        class UserList(BaseModel):\n            users: list[User]\n        ```\n        Then use `extract(\"...\", schema=UserList).users` to get it.\n\n    Examples:\n    - Correct: `extract(\"get user\", schema=User)` where `User` inherits from `BaseModel`.\n    - Incorrect: `extract(\"get user\", schema=BaseModel)` (do not pass the base type).\n\"#\n\ntemplate_string FormatHistory(history: History[]) #\"\n    {% for state in history %}\n    {% if loop.index == loop.length %}\n    Current State:\n    {% else %}\n    State ({{ loop.index0 }}):\n    {% endif %}\n        Page: {{ state.page_id or 'Unknown' }}\n    {% if state.description %}\n        Description: {{ state.description }}\n    {% endif %}\n    {% if state.layout %}\n        Layout: {{ state.layout }}\n    {% endif %}\n    {% if loop.index < loop.length %}\n        Action: {{ history[loop.index].prev_action }}\n    {% endif %}\n    {% endfor %}\n\"#\n\ntemplate_string FormatFeedback(feedback: Feedback[]) #\"\n    {% if feedback %}\n    {% for f in feedback %}\n    {{ _.role(\"user\") }}\n    # Feedback\n\n    Your previous assertion(s) might be incorrect:\n\n    {{ f.response }}\n\n    Reason: {{ f.reason }}\n\n    If you insist that it is correct, please output the same thing\n    Otherwise, modify the assertion\n    {% endfor %}\n    {% endif %}\n\"#",
    "executor/methods.baml": "class Output {\n    @@dynamic\n}\n\ntype Primitives = string | int | float | bool\ntype ExtractedData = Output | Primitives | Output[] | Primitives[]\n\n\nfunction ExtractFromState(screenshot: image, instruction: string) -> Output {\n    client \"Base\"\n    prompt #\"\n        {{_.role(\"user\")}}\n        Extract structured data from the webpage screenshot that is relevant to the instruction.\n\n        Screenshot: {{ screenshot }}\n\n        Instruction: {{ instruction }}\n        Make sure to comply the schema requirements. In case of list, no element match should return an empty list, not null.\n\n        {{ ctx.output_format }}\n    \"#\n}\n\nfunction ExtractFromElement(screenshot: image, html: string, instruction: string) -> Output {\n    client \"Base\"\n    prompt #\"\n        {{_.role(\"user\")}}\n        Extract structured data from the web UI element screenshot and HTML that is relevant to the instruction.\n\n        {{ screenshot }}\n\n        {{ html }}\n\n        Instruction: {{ instruction }}\n\n        {{ ctx.output_format }}\n    \"#\n}",
    "executor/postcondition.baml": "template_string PostconditionIntro() #\"\n    # Role\n    You are an expert QA tester.\n\n    # Objective\n    You are generating a **postcondition assertion** after a specific user action has been executed.\n    Your goal is to verify that the intended **effects** of the action have occurred.\n\n    # Instructions\n    - Construct a Python assertion function using the provided Session, State, and Element APIs as detailed below.\n    - Focus on **postcondition verification**: ensure the *intended outcome* is reflected in the state after the action.\n    - Identify which dependency types are relevant to the state change:\n        1. **Temporal Dependency:** Changes in a logical page over time (e.g., after an action, a formerly empty cart now has items).\n        2. **Data Dependency:** Propagation of information across states (e.g., product details remain consistent from search result to cart addition).\n        3. **Causal Dependency:** State changes resulting directly from user actions (e.g., clicking 'search' updates the page to show related items).\n    - Grounding: Use only information provided in the session or state. Do not invent or guess labels, text, or values.\n    - Prefer structural checks (e.g., count > 0, len >= N, is not None) when exact expected values are not known.\n    - No placeholders. Even if expectations are minimal.\n\n    - Write the assertion as a Python block:\n        ```python\n        def postcondition(session: Session):\n            ...\n        ```\n\n    # Important on assertion messages\n    - Assertions MUST have insightful, helpful messages with it, i.e. \"Message shown is not ...\", ...\n    - Messages should be helpful feedbacks for developer to know what to improve.\n\n    # MUST remember\n    - `schema` are classes that inherit from `BaseModel` but MUST MUST MUST NOT be `BaseModel` itself!.\n    - Code MUST be executable and does not have type error.\n    - Data models should be concise and sufficient for the assertions, unrelated details should be ignored.\n    - Use descriptive variable names for clarity.\n    - Minimal and to the point assertions.\n\n    # Avoid Common Pitfalls\n    - Avoid writing hardcoded values.\n    - Avoid checking page titles, `page_id` or `url`.\n    - Avoid matching exact terms or keywords in assertions, extract and check empty is better.\n\"#\n\ntemplate_string PostconditionExample() #\"\n    # Example 1\n    __input__\n    History:\n        State (0):\n            Page: Cart page;\n            Action: User clicks \"Continue Shopping\"\n        ...\n        State (4):\n            Page: Product detail view\n            Action: User adds the item to cart\n\n    Current: Cart page (After action)\n    Assert: Cart is correctly updated\n\n    __output__\n    ```python\n    def postcondition(session: Session):\n        # Define data models\n        class Product(BaseModel):\n            title: str = Field(..., description=\"The name of the product\")\n            price: float = Field(..., description=\"The unit price of the product in local currency\")\n            quantity: Optional[int] = Field(None, \"The quantity of this product (used in cart contexts). None indicates unlimited or not specified\")\n        \n        class Cart(BaseModel):\n            items: List[Product] = Field(default_factory=list, description=\"List of products in the cart with their respective quantities\")\n\n        # Extract product from latest state\n        added = session.history[-2].extract(\"get product detail\", schema=Product)\n\n        # Get current and prior cart items\n        current = session.history[-1].extract(\"get cart summary\", schema=Cart).items\n        prior = session.history[0].extract(\"get cart summary\", schema=Cart).items\n\n        # Assert cart contains prior items plus the added product\n        assert set(p.title for p in current) == set(p.title for p in prior + [added]) , \"Cart items do not match expected items after addition.\"\n\n    # Example 2: Check boolean flags\n    __input__\n    History: ...\n    Current State:\n        Page: Service Page\n        Description: ...\n        Layout: <Page>...</Page>\n    Current: Click \"Service\" button\n    Assert: Service page shows up\n\n    __output__\n    ```python\n    def postcondition(session: Session):\n        # Define data models for presence checking\n        class PresenceChecking(BaseModel):\n            service_menu_is_selected: bool = Field(..., description=\"Indicates if option A is selected\")\n            service_page_is_present: bool = Field(..., description=\"Indicates if the current page is related to services\")\n        \n        presence_checking = session.history[-1].extract(\"Check if the page is a service page\", schema=PresenceChecking)\n\n        # Check if service page is present\n        assert presence_checking.service_page_is_present is True, \"Service page is not displayed as expected.\"\n\n        # Check other flags if needed\n        assert presence_checking.service_menu_is_selected is False, \"Service menu should be selected.\"\n    ```\n\"#\n\nfunction GeneratePostcondition(screenshot: image, history: History[], action: string, verify: string, feedback: Feedback[]) -> string {\n    client \"Base\"\n    prompt #\"\n        {{_.role(\"system\")}}\n        {{ PostconditionIntro() }}\n        {{ Specification() }}\n        {{ PostconditionExample() }}\n\n        {{_.role(\"user\")}}\n        {{ screenshot }}\n\n        {{ FormatHistory(history) }}\n        After Action: {{ action }}\n        Assert: {{ verify }}\n        {{ FormatFeedback(feedback) }}\n    \"#\n}\n\n\ntest TestName {\n  functions [GeneratePostcondition]\n  args {\n    verify #\"\n      hello world\n    \"#\n    history [\n        {\n            page_id \"A\"\n        }\n    ]\n    feedback [\n        {\n            response \"test\"\n            reason \"test\"\n        }\n    ]\n    screenshot { url \"https://imgs.xkcd.com/comics/standards.png\" }\n  }\n}\n",
    "executor/precondition.baml": "template_string PreconditionIntro() #\"\n    # Role\n    You are an expert QA tester.\n\n    # Objective\n    You are generating a **precondition assertion** before a specific user action is executed.\n    Your goal is to verify that the current state is valid and ready for the intended action.\n\n    # Instructions\n    - Construct a Python assertion function using the provided Session, State, and Element APIs as detailed below.\n    - Focus on **postcondition verification**: ensure the *intended outcome* is reflected in the state after the action.\n    - Identify which dependency types are relevant to the state change:\n        1. **Temporal Dependency:** Changes in a logical page over time (e.g., after an action, a formerly empty cart now has items).\n        2. **Data Dependency:** Propagation of information across states (e.g., product details remain consistent from search result to cart addition).\n        3. **Causal Dependency:** State changes resulting directly from user actions (e.g., clicking 'search' updates the page to show related items).\n    - Grounding: Use only information provided in the session or state. Do not invent or guess labels, text, or values.\n    - Prefer structural checks (e.g., count > 0, len >= N, is not None) when exact expected values are not known.\n    - No placeholders. Even if expectations are minimal.\n    \n    - Write the assertion as a Python block:\n        ```python\n        def precondition(session: Session):\n            ...\n        ```\n\n    # Reminder\n    - `schema` must be classes that inherit from `BaseModel` but can NOT be `BaseModel`.\n\"#\n\ntemplate_string PreconditionExample() #\"\n    # Example\n    __input__\n    History:\n        State (0):\n            Page: Product search page\n            Action: User searches for 'Laptop'\n        State (1):\n            Page: Search results page\n            Action: User clicks on the desired product\n\n    Current: Product detail page (Before action)\n    Assert: Product information is loaded and 'Add to Cart' button is enabled\n\n    __output__\n    ```python\n    def precondition(session: Session):\n        # Define data model\n        class ProductDetail(BaseModel):\n            title: str = Field(..., description=\"The name of the product\")\n            price: float = Field(..., description=\"The price of the product\")\n            add_to_cart_enabled: bool = Field(..., description=\"Whether the 'Add to Cart' button is clickable\")\n\n        # Extract product details from the current page\n        details = session.history[-1].extract(\"get product detail with add-to-cart availability\", schema=ProductDetail)\n\n        # Assert that product information is complete and ready for action\n        assert details.title and details.price > 0\n        assert details.add_to_cart_enabled\n    ```\n\"#\n\nfunction GeneratePrecondition(screenshot: image, history: History[], action: string, verify: string, feedback: Feedback[]) -> string {\n    client \"Base\"\n    prompt #\"\n        {{_.role(\"system\")}}\n        {{ PreconditionIntro() }}\n        {{ Specification() }}\n        {{ PreconditionExample() }}\n        \n        {{_.role(\"user\")}}\n        {{ screenshot }}\n\n        {{ FormatHistory(history) }}\n        Before Action: {{ action }}\n        Assert: {{ verify }}\n        {{ FormatFeedback(feedback) }}\n    \"#\n}\n",
    "executor/reidentification.baml": "class PageAbstract {\n  name string\n  description string\n  layout string\n}\n\n\nfunction IsSameLogicalPage(page_a: image, page_b: image) -> bool {\n  client \"Base\"\n  prompt #\"\n    {{_.role(\"user\")}}\n    You are an expert UI/UX evaluator with deep knowledge of web page layout, semantics, and interaction patterns.\n\n    You are given two page screenshots. Determine whether they represent the same logical page \n    (i.e., the same underlying page in a web application or site), even if their visual states differ.\n\n    Guidelines:\n    1. Layout Consistency: Do the overall structure and arrangement of UI components match?\n\n    2. Functional Equivalence: Do they offer the same set of core actions and interactions?\n\n    3. Navigation Equivalence: Do they provide the same pathways to other parts of the application?\n\n    4. Two pages can be the same, even if their visual states differ. For example:\n        - One shows an expanded/triggered widget (e.g., dropdown, modal, accordion, pop-up).\n        - One contains different dynamic data but with the same type, structure, and placement.\n        - One has pre-filled or empty form fields, as long as the underlying form is the same.\n\n    {{ ctx.output_format }} with only \"True\" or \"False\".\n\n    {{ page_a }}\n\n    {{ page_b}}\n  \"#\n}\n\n\nfunction AbstractPage(page: image) -> PageAbstract {\n  client \"Base\"\n  prompt #\"\n    {{_.role(\"user\")}}\n    You are an expert UI/UX evaluator with deep knowledge of web page layout, semantics, and interaction patterns.\n    \n    You are given a page screenshot.\n    1. Assign a clear, semantically meaningful name to the page (e.g., Shopping Cart Page).\n    2. Write a summary <15 words describing both the function and visual state of the page.\n    3. Analyze the layout: extract as a page template in XML format. You may include non-data attributes for context.\n      ```example\n      <Page>\n        <Header visibleFor=\"allUsers\" role=\"navigation\" hasNotifications=\"true\" />\n        <Sidebar collapsible=\"true\" visible=\"true\" contains=\"menuItems\" />\n        <Breadcrumb currentPage=\"true\" hasPath=\"true\" />\n        <MainContent>\n          <CartItems type=\"list\" itemType=\"product\" selectable=\"true\" hasQuantity=\"true\" hasPrice=\"true\" availabilityState=\"enum\" />\n          <Recommendations type=\"list\" itemType=\"product\" algorithmType=\"collaborativeFiltering\" />\n        </MainContent>\n        <Footer role=\"footer\" contains=\"links,contactInfo\" />\n      </Page>\n      ```\n\n    {{ ctx.output_format }}\n\n    {{ page }}\n  \"#\n}\n\ntest TestName {\n  functions [AbstractPage]\n  args {\n    page { file \"/home/xiwen/WebTestPilot/webtestpilot/baml_src/executor/invoiceninja.png\" }\n  }\n}",
    "generators.baml": "// This helps use auto generate libraries you can use in the language of\n// your choice. You can have multiple generators if you use multiple languages.\n// Just ensure that the output_dir is different for each generator.\ngenerator target {\n    // Valid values: \"python/pydantic\", \"typescript\", \"ruby/sorbet\", \"rest/openapi\"\n    output_type \"python/pydantic\"\n\n    // Where the generated code will be saved (relative to baml_src/)\n    output_dir \"../src\"\n\n    // The version of the BAML package you have installed (e.g. same version as your baml-py or @boundaryml/baml).\n    // The BAML VSCode extension version should also match this version.\n    version \"0.205.0\"\n\n    // Valid values: \"sync\", \"async\"\n    // This controls what `b.FunctionName()` will be (sync or async).\n    default_client_mode sync\n}\n",
    "parser/parser.baml": "class Step {\n  condition string\n  action string\n  expectation string\n}\n\nclass TestCase {\n  name string\n  steps Step[]\n}\n\nfunction ExtractTestCase(description: string) -> TestCase {\n  client \"Base\"\n  prompt #\"\n    You are an expert in semantic intent extraction for web application testing.\n\n    Your goal is to extract structured interaction intents as a list of step = (condition, action, expectation) triples from natural language requirements or acceptance criteria.\n    - `Condition`: The context under which the step is valid (e.g., user is on page)\n    - `Action`: A clear user action (e.g., user clicks Submit)\n    - `Expectation`: A testable outcome (e.g., page redirects to dashboard)\n\n    If the information is not provided, fill with <unknown> placeholder.\n    Please correct typos and other errors to produce well-formed, complete sentences, without adding or omitting any information.\n\n    {{ ctx.output_format }}\n\n    {{ description }}\n  \"#\n}\n\nfunction ImplicitGeneration(testcase: TestCase) -> TestCase {\n  client \"Base\"\n  prompt #\"\n    You are an expert in semantic intent understanding for web application testing.\n    Fill in the <unknown> blanks by inferring contextual information from other steps.\n    Return in the same format.\n\n    {{ testcase }}\n  \"#\n}",
}

def get_baml_files():
    return _file_map


# ----------------------------------------------------------------------------
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
# ----------------------------------------------------------------------------

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code using: baml-cli generate
# baml-cli is available with the baml package.

import typing
import typing_extensions
import baml_py

from . import stream_types, types, type_builder
from .parser import LlmResponseParser, LlmStreamParser
from .runtime import DoNotUseDirectlyCallManager, BamlCallOptions
from .globals import DO_NOT_USE_DIRECTLY_UNLESS_YOU_KNOW_WHAT_YOURE_DOING_RUNTIME as __runtime__


class BamlAsyncClient:
    __options: DoNotUseDirectlyCallManager
    __stream_client: "BamlStreamClient"
    __http_request: "BamlHttpRequestClient"
    __http_stream_request: "BamlHttpStreamRequestClient"
    __llm_response_parser: LlmResponseParser
    __llm_stream_parser: LlmStreamParser

    def __init__(self, options: DoNotUseDirectlyCallManager):
        self.__options = options
        self.__stream_client = BamlStreamClient(options)
        self.__http_request = BamlHttpRequestClient(options)
        self.__http_stream_request = BamlHttpStreamRequestClient(options)
        self.__llm_response_parser = LlmResponseParser(options)
        self.__llm_stream_parser = LlmStreamParser(options)

    def with_options(self,
        tb: typing.Optional[type_builder.TypeBuilder] = None,
        client_registry: typing.Optional[baml_py.baml_py.ClientRegistry] = None,
        collector: typing.Optional[typing.Union[baml_py.baml_py.Collector, typing.List[baml_py.baml_py.Collector]]] = None,
        env: typing.Optional[typing.Dict[str, typing.Optional[str]]] = None,
    ) -> "BamlAsyncClient":
        options: BamlCallOptions = {}
        if tb is not None:
            options["tb"] = tb
        if client_registry is not None:
            options["client_registry"] = client_registry
        if collector is not None:
            options["collector"] = collector
        if env is not None:
            options["env"] = env
        return BamlAsyncClient(self.__options.merge_options(options))

    @property
    def stream(self):
      return self.__stream_client

    @property
    def request(self):
      return self.__http_request

    @property
    def stream_request(self):
      return self.__http_stream_request

    @property
    def parse(self):
      return self.__llm_response_parser

    @property
    def parse_stream(self):
      return self.__llm_stream_parser
    
    async def AbstractPage(self, page: baml_py.Image,
        baml_options: BamlCallOptions = {},
    ) -> types.PageAbstract:
        result = await self.__options.merge_options(baml_options).call_function_async(function_name="AbstractPage", args={
            "page": page,
        })
        return typing.cast(types.PageAbstract, result.cast_to(types, types, stream_types, False, __runtime__))
    async def ExtractFromElement(self, screenshot: baml_py.Image,html: str,instruction: str,
        baml_options: BamlCallOptions = {},
    ) -> types.Output:
        result = await self.__options.merge_options(baml_options).call_function_async(function_name="ExtractFromElement", args={
            "screenshot": screenshot,"html": html,"instruction": instruction,
        })
        return typing.cast(types.Output, result.cast_to(types, types, stream_types, False, __runtime__))
    async def ExtractFromState(self, screenshot: baml_py.Image,instruction: str,
        baml_options: BamlCallOptions = {},
    ) -> types.Output:
        result = await self.__options.merge_options(baml_options).call_function_async(function_name="ExtractFromState", args={
            "screenshot": screenshot,"instruction": instruction,
        })
        return typing.cast(types.Output, result.cast_to(types, types, stream_types, False, __runtime__))
    async def ExtractTestCase(self, description: str,
        baml_options: BamlCallOptions = {},
    ) -> types.TestCase:
        result = await self.__options.merge_options(baml_options).call_function_async(function_name="ExtractTestCase", args={
            "description": description,
        })
        return typing.cast(types.TestCase, result.cast_to(types, types, stream_types, False, __runtime__))
    async def GeneratePostcondition(self, screenshot: baml_py.Image,history: typing.List["types.History"],action: str,verify: str,feedback: typing.List["types.Feedback"],
        baml_options: BamlCallOptions = {},
    ) -> str:
        result = await self.__options.merge_options(baml_options).call_function_async(function_name="GeneratePostcondition", args={
            "screenshot": screenshot,"history": history,"action": action,"verify": verify,"feedback": feedback,
        })
        return typing.cast(str, result.cast_to(types, types, stream_types, False, __runtime__))
    async def GeneratePrecondition(self, screenshot: baml_py.Image,history: typing.List["types.History"],action: str,verify: str,feedback: typing.List["types.Feedback"],
        baml_options: BamlCallOptions = {},
    ) -> str:
        result = await self.__options.merge_options(baml_options).call_function_async(function_name="GeneratePrecondition", args={
            "screenshot": screenshot,"history": history,"action": action,"verify": verify,"feedback": feedback,
        })
        return typing.cast(str, result.cast_to(types, types, stream_types, False, __runtime__))
    async def ImplicitGeneration(self, testcase: types.TestCase,
        baml_options: BamlCallOptions = {},
    ) -> types.TestCase:
        result = await self.__options.merge_options(baml_options).call_function_async(function_name="ImplicitGeneration", args={
            "testcase": testcase,
        })
        return typing.cast(types.TestCase, result.cast_to(types, types, stream_types, False, __runtime__))
    async def IsSameLogicalPage(self, page_a: baml_py.Image,page_b: baml_py.Image,
        baml_options: BamlCallOptions = {},
    ) -> bool:
        result = await self.__options.merge_options(baml_options).call_function_async(function_name="IsSameLogicalPage", args={
            "page_a": page_a,"page_b": page_b,
        })
        return typing.cast(bool, result.cast_to(types, types, stream_types, False, __runtime__))
    async def LocateUIElement(self, screenshot: baml_py.Image,description: str,
        baml_options: BamlCallOptions = {},
    ) -> str:
        result = await self.__options.merge_options(baml_options).call_function_async(function_name="LocateUIElement", args={
            "screenshot": screenshot,"description": description,
        })
        return typing.cast(str, result.cast_to(types, types, stream_types, False, __runtime__))
    async def ProposeActions(self, screenshot: baml_py.Image,task: str,
        baml_options: BamlCallOptions = {},
    ) -> str:
        result = await self.__options.merge_options(baml_options).call_function_async(function_name="ProposeActions", args={
            "screenshot": screenshot,"task": task,
        })
        return typing.cast(str, result.cast_to(types, types, stream_types, False, __runtime__))
    


class BamlStreamClient:
    __options: DoNotUseDirectlyCallManager

    def __init__(self, options: DoNotUseDirectlyCallManager):
        self.__options = options

    def AbstractPage(self, page: baml_py.Image,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.BamlStream[stream_types.PageAbstract, types.PageAbstract]:
        ctx, result = self.__options.merge_options(baml_options).create_async_stream(function_name="AbstractPage", args={
            "page": page,
        })
        return baml_py.BamlStream[stream_types.PageAbstract, types.PageAbstract](
          result,
          lambda x: typing.cast(stream_types.PageAbstract, x.cast_to(types, types, stream_types, True, __runtime__)),
          lambda x: typing.cast(types.PageAbstract, x.cast_to(types, types, stream_types, False, __runtime__)),
          ctx,
        )
    def ExtractFromElement(self, screenshot: baml_py.Image,html: str,instruction: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.BamlStream[stream_types.Output, types.Output]:
        ctx, result = self.__options.merge_options(baml_options).create_async_stream(function_name="ExtractFromElement", args={
            "screenshot": screenshot,"html": html,"instruction": instruction,
        })
        return baml_py.BamlStream[stream_types.Output, types.Output](
          result,
          lambda x: typing.cast(stream_types.Output, x.cast_to(types, types, stream_types, True, __runtime__)),
          lambda x: typing.cast(types.Output, x.cast_to(types, types, stream_types, False, __runtime__)),
          ctx,
        )
    def ExtractFromState(self, screenshot: baml_py.Image,instruction: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.BamlStream[stream_types.Output, types.Output]:
        ctx, result = self.__options.merge_options(baml_options).create_async_stream(function_name="ExtractFromState", args={
            "screenshot": screenshot,"instruction": instruction,
        })
        return baml_py.BamlStream[stream_types.Output, types.Output](
          result,
          lambda x: typing.cast(stream_types.Output, x.cast_to(types, types, stream_types, True, __runtime__)),
          lambda x: typing.cast(types.Output, x.cast_to(types, types, stream_types, False, __runtime__)),
          ctx,
        )
    def ExtractTestCase(self, description: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.BamlStream[stream_types.TestCase, types.TestCase]:
        ctx, result = self.__options.merge_options(baml_options).create_async_stream(function_name="ExtractTestCase", args={
            "description": description,
        })
        return baml_py.BamlStream[stream_types.TestCase, types.TestCase](
          result,
          lambda x: typing.cast(stream_types.TestCase, x.cast_to(types, types, stream_types, True, __runtime__)),
          lambda x: typing.cast(types.TestCase, x.cast_to(types, types, stream_types, False, __runtime__)),
          ctx,
        )
    def GeneratePostcondition(self, screenshot: baml_py.Image,history: typing.List["types.History"],action: str,verify: str,feedback: typing.List["types.Feedback"],
        baml_options: BamlCallOptions = {},
    ) -> baml_py.BamlStream[str, str]:
        ctx, result = self.__options.merge_options(baml_options).create_async_stream(function_name="GeneratePostcondition", args={
            "screenshot": screenshot,"history": history,"action": action,"verify": verify,"feedback": feedback,
        })
        return baml_py.BamlStream[str, str](
          result,
          lambda x: typing.cast(str, x.cast_to(types, types, stream_types, True, __runtime__)),
          lambda x: typing.cast(str, x.cast_to(types, types, stream_types, False, __runtime__)),
          ctx,
        )
    def GeneratePrecondition(self, screenshot: baml_py.Image,history: typing.List["types.History"],action: str,verify: str,feedback: typing.List["types.Feedback"],
        baml_options: BamlCallOptions = {},
    ) -> baml_py.BamlStream[str, str]:
        ctx, result = self.__options.merge_options(baml_options).create_async_stream(function_name="GeneratePrecondition", args={
            "screenshot": screenshot,"history": history,"action": action,"verify": verify,"feedback": feedback,
        })
        return baml_py.BamlStream[str, str](
          result,
          lambda x: typing.cast(str, x.cast_to(types, types, stream_types, True, __runtime__)),
          lambda x: typing.cast(str, x.cast_to(types, types, stream_types, False, __runtime__)),
          ctx,
        )
    def ImplicitGeneration(self, testcase: types.TestCase,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.BamlStream[stream_types.TestCase, types.TestCase]:
        ctx, result = self.__options.merge_options(baml_options).create_async_stream(function_name="ImplicitGeneration", args={
            "testcase": testcase,
        })
        return baml_py.BamlStream[stream_types.TestCase, types.TestCase](
          result,
          lambda x: typing.cast(stream_types.TestCase, x.cast_to(types, types, stream_types, True, __runtime__)),
          lambda x: typing.cast(types.TestCase, x.cast_to(types, types, stream_types, False, __runtime__)),
          ctx,
        )
    def IsSameLogicalPage(self, page_a: baml_py.Image,page_b: baml_py.Image,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.BamlStream[bool, bool]:
        ctx, result = self.__options.merge_options(baml_options).create_async_stream(function_name="IsSameLogicalPage", args={
            "page_a": page_a,"page_b": page_b,
        })
        return baml_py.BamlStream[bool, bool](
          result,
          lambda x: typing.cast(bool, x.cast_to(types, types, stream_types, True, __runtime__)),
          lambda x: typing.cast(bool, x.cast_to(types, types, stream_types, False, __runtime__)),
          ctx,
        )
    def LocateUIElement(self, screenshot: baml_py.Image,description: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.BamlStream[str, str]:
        ctx, result = self.__options.merge_options(baml_options).create_async_stream(function_name="LocateUIElement", args={
            "screenshot": screenshot,"description": description,
        })
        return baml_py.BamlStream[str, str](
          result,
          lambda x: typing.cast(str, x.cast_to(types, types, stream_types, True, __runtime__)),
          lambda x: typing.cast(str, x.cast_to(types, types, stream_types, False, __runtime__)),
          ctx,
        )
    def ProposeActions(self, screenshot: baml_py.Image,task: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.BamlStream[str, str]:
        ctx, result = self.__options.merge_options(baml_options).create_async_stream(function_name="ProposeActions", args={
            "screenshot": screenshot,"task": task,
        })
        return baml_py.BamlStream[str, str](
          result,
          lambda x: typing.cast(str, x.cast_to(types, types, stream_types, True, __runtime__)),
          lambda x: typing.cast(str, x.cast_to(types, types, stream_types, False, __runtime__)),
          ctx,
        )
    

class BamlHttpRequestClient:
    __options: DoNotUseDirectlyCallManager

    def __init__(self, options: DoNotUseDirectlyCallManager):
        self.__options = options

    async def AbstractPage(self, page: baml_py.Image,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="AbstractPage", args={
            "page": page,
        }, mode="request")
        return result
    async def ExtractFromElement(self, screenshot: baml_py.Image,html: str,instruction: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="ExtractFromElement", args={
            "screenshot": screenshot,"html": html,"instruction": instruction,
        }, mode="request")
        return result
    async def ExtractFromState(self, screenshot: baml_py.Image,instruction: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="ExtractFromState", args={
            "screenshot": screenshot,"instruction": instruction,
        }, mode="request")
        return result
    async def ExtractTestCase(self, description: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="ExtractTestCase", args={
            "description": description,
        }, mode="request")
        return result
    async def GeneratePostcondition(self, screenshot: baml_py.Image,history: typing.List["types.History"],action: str,verify: str,feedback: typing.List["types.Feedback"],
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="GeneratePostcondition", args={
            "screenshot": screenshot,"history": history,"action": action,"verify": verify,"feedback": feedback,
        }, mode="request")
        return result
    async def GeneratePrecondition(self, screenshot: baml_py.Image,history: typing.List["types.History"],action: str,verify: str,feedback: typing.List["types.Feedback"],
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="GeneratePrecondition", args={
            "screenshot": screenshot,"history": history,"action": action,"verify": verify,"feedback": feedback,
        }, mode="request")
        return result
    async def ImplicitGeneration(self, testcase: types.TestCase,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="ImplicitGeneration", args={
            "testcase": testcase,
        }, mode="request")
        return result
    async def IsSameLogicalPage(self, page_a: baml_py.Image,page_b: baml_py.Image,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="IsSameLogicalPage", args={
            "page_a": page_a,"page_b": page_b,
        }, mode="request")
        return result
    async def LocateUIElement(self, screenshot: baml_py.Image,description: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="LocateUIElement", args={
            "screenshot": screenshot,"description": description,
        }, mode="request")
        return result
    async def ProposeActions(self, screenshot: baml_py.Image,task: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="ProposeActions", args={
            "screenshot": screenshot,"task": task,
        }, mode="request")
        return result
    

class BamlHttpStreamRequestClient:
    __options: DoNotUseDirectlyCallManager

    def __init__(self, options: DoNotUseDirectlyCallManager):
        self.__options = options

    async def AbstractPage(self, page: baml_py.Image,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="AbstractPage", args={
            "page": page,
        }, mode="stream")
        return result
    async def ExtractFromElement(self, screenshot: baml_py.Image,html: str,instruction: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="ExtractFromElement", args={
            "screenshot": screenshot,"html": html,"instruction": instruction,
        }, mode="stream")
        return result
    async def ExtractFromState(self, screenshot: baml_py.Image,instruction: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="ExtractFromState", args={
            "screenshot": screenshot,"instruction": instruction,
        }, mode="stream")
        return result
    async def ExtractTestCase(self, description: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="ExtractTestCase", args={
            "description": description,
        }, mode="stream")
        return result
    async def GeneratePostcondition(self, screenshot: baml_py.Image,history: typing.List["types.History"],action: str,verify: str,feedback: typing.List["types.Feedback"],
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="GeneratePostcondition", args={
            "screenshot": screenshot,"history": history,"action": action,"verify": verify,"feedback": feedback,
        }, mode="stream")
        return result
    async def GeneratePrecondition(self, screenshot: baml_py.Image,history: typing.List["types.History"],action: str,verify: str,feedback: typing.List["types.Feedback"],
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="GeneratePrecondition", args={
            "screenshot": screenshot,"history": history,"action": action,"verify": verify,"feedback": feedback,
        }, mode="stream")
        return result
    async def ImplicitGeneration(self, testcase: types.TestCase,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="ImplicitGeneration", args={
            "testcase": testcase,
        }, mode="stream")
        return result
    async def IsSameLogicalPage(self, page_a: baml_py.Image,page_b: baml_py.Image,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="IsSameLogicalPage", args={
            "page_a": page_a,"page_b": page_b,
        }, mode="stream")
        return result
    async def LocateUIElement(self, screenshot: baml_py.Image,description: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="LocateUIElement", args={
            "screenshot": screenshot,"description": description,
        }, mode="stream")
        return result
    async def ProposeActions(self, screenshot: baml_py.Image,task: str,
        baml_options: BamlCallOptions = {},
    ) -> baml_py.baml_py.HTTPRequest:
        result = await self.__options.merge_options(baml_options).create_http_request_async(function_name="ProposeActions", args={
            "screenshot": screenshot,"task": task,
        }, mode="stream")
        return result
    

b = BamlAsyncClient(DoNotUseDirectlyCallManager({}))


import ast
import re
import asyncio
from playwright.async_api import Page, ElementHandle

_current_page: Page | None = None


def _set_page(page: Page):
    global _current_page
    _current_page = page


def _require_page():
    if _current_page is None:
        raise RuntimeError("No active page. Call _set_page() first.")


def _parse_point(point_str: str) -> tuple[int, int]:
    """
    Parse string like '<point>x1 y1</point>' into (x1, y1).
    """
    match = re.match(r"<point>\s*(-?\d+)\s+(-?\d+)\s*</point>", point_str)
    if not match:
        raise ValueError(f"Invalid point format: {point_str}")
    x, y = map(int, match.groups())
    return x, y


async def _get_element(page: Page, x: int, y: int) -> ElementHandle | None:
    handle = await page.evaluate_handle(f"""
    () => {{
        const clickableTags = ['BUTTON', 'A', 'INPUT', 'TEXTAREA', 'SELECT'];
        const clickableRoles = ['button', 'link', 'menuitem'];

        const els = document.elementsFromPoint({x}, {y});
        for (const el of els) {{
            const style = window.getComputedStyle(el);
            if (
                style.pointerEvents !== 'none' &&
                style.visibility !== 'hidden' &&
                el.offsetWidth > 0 &&
                el.offsetHeight > 0 &&
                (
                    clickableTags.includes(el.tagName) ||
                    clickableRoles.includes(el.getAttribute('role')) ||
                    el.getAttribute('onclick')
                )
            ) {{
                return el;
            }}
        }}
        return els[0] || null;
    }}
    """)
    return handle.as_element() if handle else None


async def click(point: str):
    _require_page()
    x, y = _parse_point(point)
    el = await _get_element(_current_page, x, y)
    if el:
        await el.click()
    else:
        await _current_page.mouse.click(x, y)


async def left_double(point: str):
    _require_page()
    x, y = _parse_point(point)
    el = await _get_element(_current_page, x, y)
    if el:
        await el.dblclick()
    else:
        # Playwright mouse doesn't have dblclick with coordinates directly, do two clicks quickly
        await _current_page.mouse.click(x, y, click_count=2)


async def right_single(point: str):
    _require_page()
    x, y = _parse_point(point)
    el = await _get_element(_current_page, x, y)
    if el:
        await el.click(button="right")
    else:
        await _current_page.mouse.click(x, y, button="right")


async def drag(start_point: str, end_point: str):
    _require_page()
    sx, sy = _parse_point(start_point)
    ex, ey = _parse_point(end_point)

    await _current_page.mouse.move(sx, sy)
    await _current_page.mouse.down()
    # interpolate steps for smooth dragging
    steps = 10
    for i in range(1, steps + 1):
        nx = sx + (ex - sx) * i / steps
        ny = sy + (ey - sy) * i / steps
        await _current_page.mouse.move(nx, ny)
        await asyncio.sleep(0.02)
    await _current_page.mouse.up()


async def hotkey(key: str):
    _require_page()
    keys = key.lower().split()
    if not (1 <= len(keys) <= 3):
        raise ValueError("Hotkey supports 1 to 3 keys only")
    # Press keys sequentially with down and up events for modifiers
    for k in keys[:-1]:
        await _current_page.keyboard.down(k)
    await _current_page.keyboard.press(keys[-1])
    for k in reversed(keys[:-1]):
        await _current_page.keyboard.up(k)


async def type(content: str):
    _require_page()
    # The content may contain escape sequences, e.g., \n
    decoded_content = content.encode("utf-8").decode("unicode_escape")
    await _current_page.keyboard.type(decoded_content)


async def scroll(point: str | None, direction: str):
    _require_page()
    direction = direction.lower()
    if direction not in {"up", "down", "left", "right"}:
        raise ValueError("Invalid scroll direction")

    coords = None
    if point is not None:
        coords = _parse_point(point)

    await _current_page.evaluate(
        """
        ([coords, direction]) => {
          let el, scrollAmountVertical, scrollAmountHorizontal;

          if (coords) {
            const [x, y] = coords;
            el = document.elementFromPoint(x, y);
            if (!el) return;
            const rect = el.getBoundingClientRect();
            scrollAmountVertical = rect.height;
            scrollAmountHorizontal = rect.width;
          } else {
            el = window;
            scrollAmountVertical = window.innerHeight;
            scrollAmountHorizontal = window.innerWidth;
          }

          switch (direction) {
            case 'up':
              el.scrollBy(0, -scrollAmountVertical);
              break;
            case 'down':
              el.scrollBy(0, scrollAmountVertical);
              break;
            case 'left':
              el.scrollBy(-scrollAmountHorizontal, 0);
              break;
            case 'right':
              el.scrollBy(scrollAmountHorizontal, 0);
              break;
          }
        }
        """,
        [coords, direction],
    )


async def wait():
    _require_page()
    # Sleep 5 seconds
    await asyncio.sleep(5)
    # Optional: take screenshot for change detection
    # For example:
    # await _current_page.screenshot(path="wait_screenshot.png")


async def finished(content: str):
    _require_page()
    decoded_content = content.encode("utf-8").decode("unicode_escape")
    print(f"UITARS finished with content:\n{decoded_content}")


async def execute(code: str, page: Page) -> list[dict]:
    """
    Execute LLM-generated UITARS Python code blocks by exposing only the allowed functions.
    Returns an empty list for trace compatibility.
    """
    _set_page(page)

    # Create async wrapper functions
    async def _click(point: str):
        await click(point)
    
    async def _left_double(point: str):
        await left_double(point)
    
    async def _right_single(point: str):
        await right_single(point)
    
    async def _drag(start_point: str, end_point: str):
        await drag(start_point, end_point)
    
    async def _hotkey(key: str):
        await hotkey(key)
    
    async def _type(content: str):
        await type(content)
    
    async def _scroll(point: str | None, direction: str):
        await scroll(point, direction)
    
    async def _wait():
        await wait()
    
    async def _finished(content: str):
        await finished(content)

    local_vars = {
        "click": _click,
        "left_double": _left_double,
        "right_single": _right_single,
        "drag": _drag,
        "hotkey": _hotkey,
        "type": _type,
        "scroll": _scroll,
        "wait": _wait,
        "finished": _finished,
    }

    # Clean triple backticks if present
    cleaned_code = re.sub(
        r"^```(?:python)?|```$", "", code.strip(), flags=re.MULTILINE
    ).strip()
    
    # Parse the code and execute each call
    tree = ast.parse(cleaned_code)
    
    for node in tree.body:
        if isinstance(node, ast.Expr) and isinstance(node.value, ast.Call):
            func_name = node.value.func.id if isinstance(node.value.func, ast.Name) else None
            if func_name and func_name in local_vars:
                args = []
                for arg in node.value.args:
                    if isinstance(arg, ast.Constant):
                        args.append(arg.value)
                    else:
                        args.append(eval(compile(ast.Expression(arg), '<string>', 'eval')))
                
                kwargs = {}
                for kw in node.value.keywords:
                    if isinstance(kw.value, ast.Constant):
                        kwargs[kw.arg] = kw.value.value
                    else:
                        kwargs[kw.arg] = eval(compile(ast.Expression(kw.value), '<string>', 'eval'))
                
                await local_vars[func_name](*args, **kwargs)
    
    return []



import base64
import logging
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Optional
from xml.etree.ElementTree import Element as XMLElement

from baml_py import Collector, Image
from playwright.async_api import Page

from src.baml_client.async_client import b
from src.baml_client.types import History, PageAbstract, Step
from src.cache import Cache, CaptureStateCache, PageContent
from src.config import Config

from ..page_reidentification.abstract import to_xml_tree
from ..page_reidentification.accessibility import AccessibilityTree
from ..page_reidentification.distance import tree_distance
from .element import Element, ElementFactory
from .state import State, StateFactory

logger = logging.getLogger(__name__)


@dataclass
class PageSnapshot:
    xml_tree: list[XMLElement]
    elements: dict[int, Element]
    title: str
    content: str
    screenshot: str


class Session:
    """
    Manages a browser test session with state tracking capabilities.

    Args:
        page (Page): A Playwright Page instance that accesses the application under test.
            It's assumed that the application is already loaded
            and any necessary prerequisites (e.g., fixtures, authentication) have been set up.
        config (Config): Runtime configurations for this test session.

    Raises:
        AssertionError: If the provided page is not a valid Page instance or is closed.
    """

    def __init__(
        self,
        page: Page,
        config: Config,
        steps: list[Step],
        cache_path: Optional[Path] = None,
        mcp_report_progress: Optional[Callable] = None,
    ):
        assert isinstance(page, Page) and not page.is_closed()

        self.trace: list[dict] = []
        self.page: Page = page
        self.config = config

        self.collector = Collector()
        self.state_factory = StateFactory(self)
        self.element_factory = ElementFactory(self)

        self.steps = steps  # Could be empty, only for mcp parallel flow now.
        self.step_counter = 0  # Track current step number for VSCode UI

        self._history: list[State] = []
        self.cache = Cache(cache_path)

        # For mcp progress update
        self.mcp_report_progress = mcp_report_progress

        self.start_time: float = time.time()
        self.action_timestamps: list[float] = []
        self.failed_step: Optional[int] = None
        self.bug_report: Optional[Any] = None  # Store BugReport if test fails
        self.success: bool = True  # Track overall test success

    @classmethod
    async def create(
        cls,
        page: Page,
        config: Config,
        steps: list[Step] = [],
        defer_capture_state: bool = False,
        cache_path: Optional[Path] = None,
        mcp_report_progress: Optional[Callable] = None,
    ) -> "Session":
        """
        Factory method to create and initialize a Session.

        Args:
            page (Page): A Playwright Page instance that accesses the application under test.
            config (Config): Runtime configurations for this test session.
            defer_capture (bool): If True, skip capturing the initial state.
            cache_path (Optional[Path]): Path to cache file for this test.

        Returns:
            Session: A fully initialized Session instance.
        """
        session = cls(page, config, steps, cache_path, mcp_report_progress)

        # --- Helper to keep following new tabs opened from the current page ---
        async def attach_popup_handler(current_page: Page):
            async def on_popup(popup: Page):
                await popup.wait_for_load_state()
                await popup.bring_to_front()
                session.page = popup  # now all further actions use the new tab

                # Get the new tab's target ID for screencast streaming
                try:
                    cdp = await popup.context.new_cdp_session(popup)
                    target_info = await cdp.send("Target.getTargetInfo")
                    new_target_id = target_info["targetInfo"]["targetId"]
                    # Log in a format the VS Code extension can parse
                    logger.info(f"NEW_TAB_OPENED: {new_target_id}")
                    # print(f"New tab opened: {popup.url} (targetId: {new_target_id})")
                except Exception as e:
                    logger.warning(f"Failed to get target ID for new tab: {e}")
                    # print(f"New tab opened: {popup.url}")

                # Also attach handler to this new page, in case it opens another tab.
                await attach_popup_handler(popup)

                # Close previous page after delay to avoid race conditions with screenshots
                await popup.wait_for_timeout(5000)
                await current_page.close()

            current_page.on("popup", on_popup)

        await attach_popup_handler(session.page)

        if not defer_capture_state:
            page_snapshot = await session.get_current_page_snapshot()
            await session.capture_state(prev_action=None, page_snapshot=page_snapshot)
        return session

    @property
    def history(self) -> list[State]:
        """
        Get the chronological history of captured states.

        Returns:
            list[State]: A read-only copy of all previously captured states in the test session,
                    ordered chronologically from oldest to newest.
        """
        return self._history.copy()

    async def update_mcp_progress(self, message: str) -> None:
        if not self.mcp_report_progress:
            return

        await self.mcp_report_progress(self.step_counter, len(self.steps), message)

    async def capture_state(
        self,
        prev_action: str | None,
        page_snapshot: PageSnapshot,
        step_number: Optional[int] = None,
    ) -> None:
        """
        Capture the current state of the browser page after an action.
        NOTE: Page snapshot is done outside and before calling this function.
        NOTE: to avoid race conditions due to page changes (in parallel action execution).
        NOTE: step_number: Optional explicit step number for cache access (for parallel execution).
        """
        step_num = step_number if step_number is not None else self.step_counter

        await self.update_mcp_progress(f"Capturing state step {step_number}...")

        cached = self.cache.get_capture_state(step_num, prev_action)
        if cached:
            page_id = cached.page_content.name
            description = cached.page_content.description
            layout = cached.page_content.layout
            logger.info(f"Cache hit for capture_state at step {step_num}")
        else:
            # Slow, blocking.
            page_id, description, layout = await self._page_reidentification(
                page_snapshot.xml_tree, page_snapshot.screenshot
            )

        # Save to cache
        self.cache.set_capture_state(
            step_num,
            CaptureStateCache(
                screenshot=page_snapshot.screenshot,
                page_content=PageContent(
                    name=page_id, description=description, layout=layout
                ),
                prev_action=prev_action,
            ),
        )

        # Update with new state
        state = self.state_factory.create(
            page_id=page_id,
            description=description,
            layout=layout,
            url=self.page.url,
            title=page_snapshot.title,
            content=page_snapshot.content,
            screenshot=page_snapshot.screenshot,
            elements=page_snapshot.elements,
            prev_action=prev_action,
            xml_tree=page_snapshot.xml_tree,
        )

        for e in page_snapshot.elements.values():
            e.state = state

        self._history.append(state)

    async def get_current_page_snapshot(self):
        tree = await AccessibilityTree.create(self.page)
        xml_tree = to_xml_tree(tree)
        elements = await self.capture_elements()
        page_title = await self.page.title()
        page_content = await self.page.content()
        screenshot = base64.b64encode(await self.page.screenshot(type="png")).decode(
            "utf-8"
        )
        return PageSnapshot(
            xml_tree=xml_tree,
            elements=elements,
            title=page_title,
            content=page_content,
            screenshot=screenshot,
        )

    async def capture_elements(self) -> dict[int, Element]:
        def _build_tree(
            elements_data: list[dict[str, Any]],
        ) -> tuple[dict[str, Element], Element]:
            elements: dict[str, Element] = {
                data["id"]: self.element_factory.create(data) for data in elements_data
            }
            root = None
            for el in elements.values():
                if el.parentId is not None:
                    parent = elements.get(el.parentId)
                    if parent:
                        parent.children.append(el)
                else:
                    root = el
            return elements, root  # type: ignore  # type: ignore

        elements_data = await self.page.evaluate("""
            (() => {
                let idCounter = 1;
                const nodes = [];

                function traverse(node, parentId = null) {
                    const id = idCounter++;
                    const rect = node.getBoundingClientRect();
                    const style = window.getComputedStyle(node);

                    const attributes = {};
                    for (const attr of node.attributes) {
                        attributes[attr.name] = attr.value;
                    }

                    nodes.push({
                        id,
                        parentId,
                        tagName: node.tagName,
                        outerHTML: node.outerHTML,
                        x: rect.x,
                        y: rect.y,
                        width: rect.width,
                        height: rect.height,
                        zIndex: parseInt(style.zIndex) || 0,
                        visible: style.visibility !== 'hidden' && style.display !== 'none' && style.opacity !== '0',
                        attributes,
                        textContent: node.textContent.trim() || null
                    });

                    for (const child of node.children) {
                        traverse(child, id);
                    }
                }

                traverse(document.documentElement, null);
                return nodes;
            })()
        """)

        elements, _ = _build_tree(elements_data)
        return elements

    def get_history(
        self,
    ) -> list[History]:
        seen_pages = set()
        history = []
        for state in self.history:
            layout = state.layout if state.page_id not in seen_pages else None
            description = state.description if state.page_id not in seen_pages else None
            history.append(
                History(
                    page_id=state.page_id,
                    layout=layout,
                    description=description,
                    prev_action=state.prev_action,
                )
            )
            seen_pages.add(state.page_id)
        return history

    async def _page_reidentification(
        self, xml_tree: list[XMLElement], screenshot: str
    ) -> tuple[str, str, str]:
        """
        Determine if the current page matches any previously visited logical page.
        If matched, return the existing page ID and description.
        Otherwise, generate a new page ID and description.

        Returns:
            tuple[str, str]: A tuple containing:
                - page_id: A short identifier or name of the logical page.
                - description: A detailed textual description of the page.
        """
        current_img = Image.from_base64("image/png", screenshot)

        # Handle empty history — no page to compare
        if not self.history:
            logger.debug("Abstracting page...")
            page_abstract: PageAbstract = await b.AbstractPage(
                current_img,
                baml_options={
                    "client_registry": self.config.page_reidentification,
                    "collector": self.collector,
                },
            )
            return page_abstract.name, page_abstract.description, page_abstract.layout

        # Find the history state with the smallest tree distance to current page
        closest_state = min(
            self.history, key=lambda s: tree_distance(xml_tree, s.xml_tree)
        )
        closest_img = Image.from_base64("image/png", closest_state.screenshot)

        logger.info("Checking page re-identification...")
        if await b.IsSameLogicalPage(
            current_img,
            closest_img,
            baml_options={
                "client_registry": self.config.page_reidentification,
                "collector": self.collector,
            },
        ):
            return (
                closest_state.page_id,
                closest_state.description,
                closest_state.layout,
            )

        logger.debug("Abstracting page...")
        new_page_abstract: PageAbstract = await b.AbstractPage(
            current_img,
            baml_options={
                "client_registry": self.config.page_reidentification,
                "collector": self.collector,
            },
        )
        return (
            new_page_abstract.name,
            new_page_abstract.description,
            new_page_abstract.layout,
        )



# ----------------------------------------------------------------------------
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
# ----------------------------------------------------------------------------

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code using: baml-cli generate
# baml-cli is available with the baml package.

import typing
import typing_extensions
from pydantic import BaseModel, ConfigDict

import baml_py

from . import types

StreamStateValueT = typing.TypeVar('StreamStateValueT')
class StreamState(BaseModel, typing.Generic[StreamStateValueT]):
    value: StreamStateValueT
    state: typing_extensions.Literal["Pending", "Incomplete", "Complete"]
# #########################################################################
# Generated classes (6)
# #########################################################################

class Feedback(BaseModel):
    response: typing.Optional[str] = None
    reason: typing.Optional[str] = None

class History(BaseModel):
    page_id: typing.Optional[str] = None
    layout: typing.Optional[str] = None
    description: typing.Optional[str] = None
    prev_action: typing.Optional[str] = None

class Output(BaseModel):
    model_config = ConfigDict(extra='allow')

class PageAbstract(BaseModel):
    name: typing.Optional[str] = None
    description: typing.Optional[str] = None
    layout: typing.Optional[str] = None

class Step(BaseModel):
    condition: typing.Optional[str] = None
    action: typing.Optional[str] = None
    expectation: typing.Optional[str] = None

class TestCase(BaseModel):
    name: typing.Optional[str] = None
    steps: typing.List["Step"]

# #########################################################################
# Generated type aliases (2)
# #########################################################################


ExtractedData: typing_extensions.TypeAlias = typing.Optional[typing.Union["Output", str, int, float, bool, typing.List["Output"], typing.List[typing.Union[str, int, float, bool]]]]


Primitives: typing_extensions.TypeAlias = typing.Optional[typing.Union[str, int, float, bool]]



from __future__ import annotations

import base64
import io
import logging
from typing import TYPE_CHECKING, Any, Optional

import PIL.Image
from baml_py import ClientRegistry, Collector, Image

from src.baml_client.sync_client import b
from src.baml_client.type_builder import TypeBuilder
from src.baml_client.types import ExtractedData

from .pydantic_schema import build_from_pydantic
from .type_utils import convert_extracted_data

if TYPE_CHECKING:
    from .session import Session
    from .state import State


logger = logging.getLogger(__name__)


class Element:
    """
    Represents a DOM-like element with spatial coordinates, dimensions, z-index, and hierarchy.
    """

    def __init__(
        self,
        data: dict[str, Any],
        client_registry: ClientRegistry,
        collector: Collector,
    ):
        self.id: int = data["id"]
        self.parentId: Optional[int] = data["parentId"]
        self.tagName: str = data["tagName"]
        self.outerHTML: str = data["outerHTML"]
        self.x: float = data["x"]
        self.y: float = data["y"]
        self.width: float = data["width"]
        self.height: float = data["height"]
        self.z_index: int = data.get("zIndex", 0) or 0
        self.visible: bool = data.get("visible", True)
        self.text_content: str = data.get("textContent", "")
        self.attributes: dict = data.get("attributes", {})

        self.children: list["Element"] = []
        self.parent: Optional["Element"] = None

        self.client_registry = client_registry
        self.collector = collector
        self.state: Optional["State"] = None

    def contains(self, px: float, py: float) -> bool:
        """
        Check if a given point lies within this element's bounding box.
        """
        return (
            self.visible
            and self.x <= px <= self.x + self.width
            and self.y <= py <= self.y + self.height
        )

    def extract(self, instruction: str, schema: ExtractedData) -> ExtractedData:
        """
        Extract structured data from the element using a schema.

        Args:
            instruction (str):
                A natural language description of information to extract.
                Example: `"get product detail"` or `"extract cart summary"`.
            schema (ExtractedData):
                A BaseModel class, primitive type, or collection type defining the expected output.

        Returns:
            ExtractedData:
                An instance of the provided `schema` type containing validated extracted data.

        Example:
            >>> class Product(BaseModel):
            ...     title: str
            ...     price: float
            ...
            >>> data = element.extract("get product detail", schema=Product)
            >>> text = element.extract("get visible text", schema=str)
            >>> items = element.extract("get all items", schema=list)
            >>> data
            {'title': 'Sample Item', 'price': 9.99}
        """
        tb = TypeBuilder()
        field_type = build_from_pydantic(schema, tb)
        tb.Output.add_property("schema", field_type)

        # Decode screenshot from base64
        image_bytes = base64.b64decode(self.state.screenshot)
        image = PIL.Image.open(io.BytesIO(image_bytes))

        # Ensure crop bounding box is within image bounds
        img_width, img_height = image.size
        x1 = max(0, min(self.x, img_width))
        y1 = max(0, min(self.y, img_height))
        x2 = max(0, min(self.x + self.width, img_width))
        y2 = max(0, min(self.y + self.height, img_height))

        crop_box = (x1, y1, x2, y2)
        cropped_image = image.crop(crop_box)

        # Convert cropped image back to base64
        buffered = io.BytesIO()
        cropped_image.save(buffered, format="PNG")
        cropped_b64 = base64.b64encode(buffered.getvalue()).decode("utf-8")
        screenshot = Image.from_base64("image/png", cropped_b64)

        output = b.ExtractFromElement(
            screenshot,
            self.outerHTML,
            instruction,
            baml_options={
                "tb": tb,
                "client_registry": self.client_registry,
                "collector": self.collector,
            },
        )
        data = convert_extracted_data(schema, output)
        logger.info(f"Extracted data: {data}")
        return data


class ElementFactory:
    def __init__(self, session: "Session"):
        self.client_registry: ClientRegistry = session.config.assertion_api
        self.collector: Collector = session.collector

    def create(self, data: dict[str, Any]):
        return Element(data, self.client_registry, self.collector)



import asyncio
import logging
import traceback
from pathlib import Path
from typing import Callable, Optional

from playwright.async_api import Page

from .baml_client.types import Step
from .config import Config
from .executor import (
    BugReport,
    execute_action,
    verify_postcondition,
)
from .executor.assertion_api import Session
from .parser import parse

logger = logging.getLogger(__name__)
Hook = Callable[[BugReport], None]


class WebTestPilot:
    @staticmethod
    def parse(config: Config, description: str) -> list["Step"]:
        test_case = parse(description, config)
        return test_case.steps

    @staticmethod
    async def run(
        session: Session,
        steps: list["Step"],
        assertion: bool,
        hooks: Optional[list[Hook]] = None,
    ) -> Session:
        """
        This is used in baseline runners as well, sequential execution of steps.
        Session here is initialized from outside.
        Execute a test case on the given Session.

        Params:
            session: The current test session.
            test_input: Description string, a single Step, or list of Steps.
            hooks: Optional list of hooks to trigger (Callables) when a BugReport occurs.

        Returns:
            Session: The session object with success status and bug_report if failed.
        """
        assert isinstance(steps, list)
        assert all(isinstance(s, Step) for s in steps)

        config = session.config
        hooks = hooks or []

        for step in steps:
            try:
                await execute_action(session, step.action, config)
                page_snapshot = await session.get_current_page_snapshot()
                await session.capture_state(
                    prev_action=step.action, page_snapshot=page_snapshot
                )
                if assertion and step.expectation:
                    await verify_postcondition(
                        session, step.action, step.expectation, config
                    )

            except BugReport as report:
                logger.error(f"Bug reported: {str(report)}")
                session.success = False
                session.bug_report = report
                for hook in hooks:
                    try:
                        hook(report)
                    except Exception:
                        logger.error("Exception in hook:", traceback.format_exc())
                return session
            except Exception as e:
                logger.error("Exception in test session:", traceback.format_exc())
                session.success = False
                session.bug_report = e
                return session

        session.cache.save()
        return session

    @staticmethod
    async def run_parallel(
        page: Page,
        config: Config,
        steps: list["Step"],
        assertion: bool,
        hooks: Optional[list[Hook]] = None,
        cache_path: Optional[Path] = None,
        mcp_report_progress: Optional[Callable] = None,
    ) -> Session:
        """
        Main differences:
        - Session is initialized in this function.
        - These steps are in parallel: verify_postcondition (current step) // execute_action (next step)
        """
        assert isinstance(steps, list)
        assert all(isinstance(s, Step) for s in steps)
        assert len(steps) >= 1, "At least one step is required"

        hooks = hooks or []

        # Parallel strategy:
        # - Step 1: Initialize session (and capture initial state) + first action (before any steps)
        # - Step 2: The rest: capture_state & verify_postcondition (current step) // execute_action (next step)

        # Step 1: Initialize session + first action in parallel
        session = await Session.create(
            page,
            config,
            steps,
            defer_capture_state=True,
            cache_path=cache_path,
            mcp_report_progress=mcp_report_progress,
        )

        page_snapshot = await session.get_current_page_snapshot()
        capture_task = asyncio.create_task(
            session.capture_state(
                prev_action=None, page_snapshot=page_snapshot, step_number=0
            )
        )
        first_action_task = asyncio.create_task(
            execute_action(session, steps[0].action, config, step_number=1)
        )

        await capture_task
        await first_action_task

        # Step 2. The rest: capture_state & verify_postcondition (current step) // execute_action (next step)
        for step_idx, (step, next_step) in enumerate(
            zip(steps, steps[1:] + [None]), start=1
        ):
            try:
                # HACK: Ensure page is stable (some components & data render slow esp on page changes)
                try:
                    await session.page.wait_for_load_state("networkidle", timeout=5000)
                except Exception:
                    pass

                # Lock the snapshot to be before (action execution + verification) so they both look at same state snapshot.
                page_snapshot = await session.get_current_page_snapshot()
                current_step_num = step_idx
                next_step_num = step_idx + 1

                async def capture_and_verify(step: Step, step_num: int):
                    # Verify always after capture.
                    await session.capture_state(
                        prev_action=step.action,
                        page_snapshot=page_snapshot,
                        step_number=step_num,
                    )
                    if assertion and step.expectation:
                        await verify_postcondition(
                            session,
                            step.action,
                            step.expectation,
                            config,
                            step_number=step_num,
                        )

                # Step N's (capture -> verify) // Step N+1's action (if present)
                verify_task = asyncio.create_task(
                    capture_and_verify(step, current_step_num)
                )
                action_task: asyncio.Task[None] | None = (
                    asyncio.create_task(
                        execute_action(
                            session, next_step.action, config, step_number=next_step_num
                        )
                    )
                    if next_step
                    else None
                )

                # Collect both results
                await verify_task
                if action_task:
                    await action_task
                    
                session.cache.save_at_step(current_step_num)
            except BugReport as report:
                logger.error(f"Bug reported: {str(report)}")
                session.success = False
                session.bug_report = report
                for hook in hooks:
                    try:
                        hook(report)
                    except Exception:
                        logger.error("Exception in hook:", traceback.format_exc())
                return session
            except Exception as e:
                logger.error("Exception in test session:", traceback.format_exc())
                session.success = False
                session.bug_report = e
                return session

        session.cache.save()
        return session


if __name__ == "__main__":
    pass



import ast
import asyncio
import base64
from io import BytesIO
import logging
import re
from copy import deepcopy
from typing import Any

from src.baml_client.async_client import b
from baml_py import ClientRegistry, Collector, Image
from PIL import ImageDraw
from PIL import Image as PILImage
from ..assertion_api.session import Session
from playwright.async_api import ElementHandle, Page

# -------------------------
# Globals
# -------------------------
logger = logging.getLogger(__name__)
_current_session: Session | None = None
_trace: list[dict] = []


def _get_current_page() -> Page:
    """Get the current page from session, allowing dynamic updates (e.g., popups)."""
    if _current_session is None:
        raise RuntimeError("No active session. Call _set_session() first.")
    return _current_session.page


# -------------------------
# Helpers
# -------------------------
async def write_image_for_debug(path: str, x: int, y: int):
    # Annotate the screenshot and write to file for debugging
    page = _get_current_page()
    with open(
        path,
        "wb",
    ) as f:
        image_b64 = base64.b64encode(await page.screenshot(type="png")).decode("utf-8")
        # draw circle at (x,y)
        image = PILImage.open(BytesIO(base64.b64decode(image_b64)))
        ImageDraw.Draw(image).ellipse(
            (x - 5, y - 5, x + 5, y + 5),
            # outline="red",
            fill="red",
            width=4,
        )
        image.save(f, format="PNG")


def _parse_coordinates(output_text: str) -> tuple[float, float]:
    # Note: this assumes screenshot is 1920 * 1080px
    # Example: output_text = "[1130,534,1553,560]"
    box = eval(output_text)

    # NOTE: This is different because model's patch is 14x14
    input_height = 1092
    input_width = 1932
    abs_x1 = float(box[0]) / input_width
    abs_y1 = float(box[1]) / input_height
    abs_x2 = float(box[2]) / input_width
    abs_y2 = float(box[3]) / input_height
    bbox = [abs_x1, abs_y1, abs_x2, abs_y2]
    point = [(bbox[0] + bbox[2]) / 2, (bbox[1] + bbox[3]) / 2]
    return point[0] * 1920, point[1] * 1080


async def _get_element(
    page: Page, x: int, y: int, target_description: str
) -> ElementHandle | None:
    """Get element at coordinates, with special handling for ul/ol to find best matching li."""
    handle = await page.evaluate_handle(f"""
    () => {{
      const clickableTags = ['BUTTON','A','INPUT','TEXTAREA','SELECT'];
      const clickableRoles = ['button','link','menuitem'];

      // create a temporary highlight circle at (x,y)
      try {{
        const highlightId = '__wpt_highlight_circle';
        const existing = document.getElementById(highlightId);
        if (existing) existing.remove();

        const r = 35; // radius in px
        const div = document.createElement('div');
        div.id = highlightId;
        Object.assign(div.style, {{
          position: 'fixed',
          left: ({x} - r) + 'px',
          top: ({y} - r) + 'px',
          width: (r * 2) + 'px',
          height: (r * 2) + 'px',
          borderRadius: '50%',
          background: 'rgba(255,200,0,0.20)',
          boxShadow: '0 0 12px 4px rgba(255,200,0,0.6)',
          border: '2px solid rgba(255,165,0,0.95)',
          pointerEvents: 'none',
          zIndex: '2147483647',
          transition: 'opacity 0.45s ease-out, transform 0.45s ease-out',
          transform: 'scale(1)'
        }});
        document.body.appendChild(div);
        // fade and remove after a short delay
        setTimeout(() => {{
          div.style.opacity = '0';
          div.style.transform = 'scale(1.4)';
          setTimeout(() => div.remove(), 450);
        }}, 600);
      }} catch (e) {{
        // ignore highlight errors
      }}

      const els = document.elementsFromPoint({x}, {y});
      for (const el of els) {{
        const style = window.getComputedStyle(el);
        if (
          style.pointerEvents !== 'none' &&
          style.visibility !== 'hidden' &&
          el.offsetWidth > 0 &&
          el.offsetHeight > 0 &&
          (clickableTags.includes(el.tagName) ||
           clickableRoles.includes(el.getAttribute('role')) ||
           el.getAttribute('onclick'))
        ) {{
          return el;
        }}
      }}
      return els[0] || null; // fallback
    }}
    """)

    if not handle:
        logger.debug(f"No element handle found at ({x}, {y})")
        return None

    logger.debug(
        f"Element handle at ({x}, {y}): {handle} {handle.as_element()} {await handle.get_properties()}"
    )
    element = handle.as_element()

    if element:
        logger.debug(
            f"Element inner html {await element.inner_html()} {await element.get_properties()}"
        )

        # Handle ul/ol by finding best matching li if target_description is provided
        tag_name = await element.evaluate("el => el.tagName.toLowerCase()")
        if tag_name in ["ul", "ol"]:
            logger.debug(
                f"Detected {tag_name} element, searching for best matching <li>"
            )
            best_li = await _find_best_matching_li(element, target_description)
            if best_li:
                element = best_li
                logger.debug("Switched to best matching <li> element")

        logger.debug(f"Element tag name: {tag_name}")
        if tag_name in ["td"]:
            logger.debug(
                f"Detected {tag_name} element, searching for best matching <a>"
            )
            best_a = await _find_best_matching_a(element, target_description)
            if best_a:
                element = best_a
                logger.debug("Switched to best matching <a> element")

        return element

    return None


async def playwright_click(get_by: str, **kwargs):
    page = _get_current_page()
    if get_by == "selector":
        await page.click(**kwargs)
    if get_by == "text":
        await page.get_by_text(**kwargs).click()
    if get_by == "role":
        await page.get_by_role(**kwargs).click()
    if get_by == "label":
        await page.get_by_label(**kwargs).click()
    if get_by == "placeholder":
        await page.get_by_placeholder(**kwargs).click()
    if get_by == "alt_text":
        await page.get_by_alt_text(**kwargs).click()
    if get_by == "title":
        await page.get_by_title(**kwargs).click()

    # playwright_click("selector", selector="#base_bd > div.main.layoutfix > div.mb10.bs06.nobs1.tb_cner > table > tbody > tr > td:nth-child(7) > a.act_del_s")
    # playwright_click("table").get_by_role("link", name="删除").click()
    # playwright_click("button", name="确认").click()



def _simple_text_similarity(text1: str, text2: str) -> float:
    """Simple text similarity based on common word overlap."""
    words1 = set(text1.lower().split())
    words2 = set(text2.lower().split())
    if not words1 or not words2:
        return 0.0
    intersection = words1.intersection(words2)
    return len(intersection) / max(len(words1), len(words2))


async def _find_best_matching_li(
    element: ElementHandle, target_description: str
) -> ElementHandle | None:
    """Find the <li> child with the most matching text when parent is <ul> or <ol>."""
    try:
        li_elements = await element.query_selector_all("li")
        if not li_elements:
            return None

        best_match, best_score = None, -1.0
        for li in li_elements:
            score = _simple_text_similarity(await li.inner_text(), target_description)
            if score > best_score:
                best_match, best_score = li, score

        return best_match
    except Exception as e:
        logger.warning(f"Error finding best matching li: {e}")
        return None


async def _find_best_matching_a(
    element: ElementHandle, target_description: str
) -> ElementHandle | None:
    """Find the <li> child with the most matching text when parent is <ul> or <ol>."""
    try:
        li_elements = await element.query_selector_all("a")
        if not li_elements:
            return None

        best_match, best_score = None, -1.0
        for li in li_elements:
            score = _simple_text_similarity(await li.inner_text(), target_description)
            if score > best_score:
                best_match, best_score = li, score

        return best_match
    except Exception as e:
        logger.warning(f"Error finding best matching li: {e}")
        return None


async def _get_xpath(element: ElementHandle) -> str:
    return await element.evaluate("""
    (el) => {
        if (!el) return '';
        const parts = [];
        while (el && el.nodeType === Node.ELEMENT_NODE) {
            let idx = 1;
            let sibling = el.previousElementSibling;
            while (sibling) {
                if (sibling.tagName === el.tagName) idx++;
                sibling = sibling.previousElementSibling;
            }
            parts.unshift(el.tagName.toLowerCase() + '[' + idx + ']');
            el = el.parentElement;
        }
        return '/' + parts.join('/');
    }
    """)


async def _get_screenshot() -> Image:
    page = _get_current_page()
    screenshot = base64.b64encode(await page.screenshot(type="png")).decode("utf-8")
    return Image.from_base64("image/png", screenshot)


def _set_session(session: Session):
    global _current_session
    _current_session = session


def _require_page():
    if _current_session is None:
        raise RuntimeError("No active session. Call _set_session() first.")

    page = _get_current_page()
    assert page is not None and not page.is_closed(), (
        "Current page is not set or closed"
    )


async def _focus(x: int, y: int) -> tuple[int, int]:
    """Make (x, y) center of viewport as much as possible.

    Returns:
        tuple[int, int]: (deltaX, deltaY) - the amount by which the coordinates shifted after scrolling
    """
    page = _get_current_page()
    result = await page.evaluate(
        """({ x, y }) => {
            const viewportWidth = window.innerWidth;
            const viewportHeight = window.innerHeight;
            const docWidth = document.documentElement.scrollWidth;
            const docHeight = document.documentElement.scrollHeight;

            const initialScrollX = window.scrollX;
            const initialScrollY = window.scrollY;

            if (docWidth <= viewportWidth && docHeight <= viewportHeight) {
                // No scrolling possible
                return { deltaX: 0, deltaY: 0 };
            }

            let scrollX = x - viewportWidth / 2;
            let scrollY = y - viewportHeight / 2;

            const maxScrollX = docWidth - viewportWidth;
            const maxScrollY = docHeight - viewportHeight;
            scrollX = Math.max(0, Math.min(scrollX, maxScrollX));
            scrollY = Math.max(0, Math.min(scrollY, maxScrollY));

            window.scrollTo(scrollX, scrollY);
            
            const deltaX = window.scrollX - initialScrollX;
            const deltaY = window.scrollY - initialScrollY;
            
            return { deltaX, deltaY };
        }""",
        {"x": x, "y": y},
    )

    logger.debug(
        f"Scrolled to focus on ({x}, {y}), delta: ({result['deltaX']}, {result['deltaY']})"
    )
    return result["deltaX"], result["deltaY"]


async def click(cr: ClientRegistry, collector: Collector, target_description: str):
    _require_page()

    screenshot = await _get_screenshot()
    logger.debug(f"Locating element to click: {target_description}")
    coordinates = await b.LocateUIElement(
        screenshot,
        target_description,
        baml_options={"client_registry": cr, "collector": collector},
    )
    x, y = _parse_coordinates(coordinates)

    # Leads to scroll -> needs to update new coordinates
    delta_x, delta_y = await _focus(x, y)
    x -= delta_x
    y -= delta_y

    page = _get_current_page()
    element: ElementHandle = await _get_element(page, x, y, target_description)
    xpath = await _get_xpath(element)
    logger.debug(
        f"Clicking element at ({x}, {y}) with element: {element} and xpath: {xpath}"
    )
    await element.click(force=True)

    _trace.append({"action": {"name": "click", "args": {"xpath": xpath}}})


async def type(
    cr: ClientRegistry, collector: Collector, target_description: str, content: str
):
    _require_page()

    screenshot = await _get_screenshot()
    coordinates = await b.LocateUIElement(
        screenshot,
        target_description,
        baml_options={"client_registry": cr, "collector": collector},
    )
    x, y = _parse_coordinates(coordinates)

    delta_x, delta_y = await _focus(x, y)
    x -= delta_x
    y -= delta_y

    page = _get_current_page()
    element: ElementHandle = await _get_element(page, x, y, target_description)

    try:
        await element.fill(content)
    except Exception as e:
        logger.warning(f"Exception during fill: {str(e)}")
        if len(content) < 25:
            await element.type(content, delay=150)
        else:
            await element.type(content)

    _trace.append(
        {"action": {"name": "fill", "args": {"xpath": await _get_xpath(element)}}}
    )


async def drag(
    cr: ClientRegistry,
    collector: Collector,
    source_description: str,
    target_description: str,
):
    _require_page()
    page = _get_current_page()

    screenshot = await _get_screenshot()
    source_coordinates = await b.LocateUIElement(
        screenshot,
        source_description,
        baml_options={"client_registry": cr, "collector": collector},
    )
    source_x, source_y = _parse_coordinates(source_coordinates)
    target_coordinates = await b.LocateUIElement(
        screenshot,
        target_description,
        baml_options={"client_registry": cr, "collector": collector},
    )
    target_x, target_y = _parse_coordinates(target_coordinates)

    # Move to source position
    await page.mouse.move(source_x, source_y)

    # Press mouse button down
    await page.mouse.down()

    # Move to target position with intermediate steps
    steps = 10
    delay_ms = 200
    for i in range(1, steps + 1):
        ratio = i / steps
        intermediate_x = int(source_x + (target_x - source_x) * ratio)
        intermediate_y = int(source_y + (target_y - source_y) * ratio)
        await page.mouse.move(intermediate_x, intermediate_y)
        await asyncio.sleep(delay_ms / 1000)

    # Move to final target position
    await page.mouse.move(target_x, target_y)

    # Move again to ensure dragover events are properly triggered
    await page.mouse.move(target_x, target_y)

    # Release mouse button
    await page.mouse.up()

    _trace.append(
        {
            "action": {
                "name": "drag",
                "args": {
                    "source_xpath": await _get_xpath(
                        await _get_element(page, source_x, source_y, source_description)
                    ),
                    "target_xpath": await _get_xpath(
                        await _get_element(page, target_x, target_y, target_description)
                    ),
                },
            }
        }
    )


async def scroll(
    cr: ClientRegistry,
    collector: Collector,
    target_description: str | None,
    direction: str,
):
    _require_page()

    direction = direction.lower()
    if direction not in {"up", "down", "left", "right"}:
        raise ValueError("direction must be one of 'up', 'down', 'left', or 'right'")

    # Default values: scroll window
    coords = None
    screenshot = await _get_screenshot()
    if target_description is not None:
        coordinates = await b.LocateUIElement(
            screenshot,
            target_description,
            baml_options={"client_registry": cr, "collector": collector},
        )
        x, y = _parse_coordinates(coordinates)
        coords = [x, y]

        page = _get_current_page()
        await page.evaluate(
            """
        ([coords, direction]) => {
        let el, scrollAmountVertical, scrollAmountHorizontal;

        if (coords) {
            const [x, y] = coords;
            el = document.elementFromPoint(x, y);
            if (!el) return;
            const rect = el.getBoundingClientRect();
            scrollAmountVertical = rect.height;
            scrollAmountHorizontal = rect.width;
        } else {
            el = window;
            scrollAmountVertical = window.innerHeight;
            scrollAmountHorizontal = window.innerWidth;
        }

        switch (direction) {
            case 'up':
            el.scrollBy(0, -scrollAmountVertical);
            break;
            case 'down':
            el.scrollBy(0, scrollAmountVertical);
            break;
            case 'left':
            el.scrollBy(-scrollAmountHorizontal, 0);
            break;
            case 'right':
            el.scrollBy(scrollAmountHorizontal, 0);
            break;
        }
        }
        """,
            [coords, direction],
        )


async def wait(duration: int):
    if duration <= 0:
        raise ValueError("Wait duration must be >0 miliseconds")

    await asyncio.sleep(duration / 1000)


async def finished():
    pass


async def execute(code: str, page: Page, session: Session) -> list[dict]:
    """
    Safely execute LLM-generated Python code blocks containing only automation actions.
    Automatically sets the current Playwright Page before execution.
    """
    cr: ClientRegistry = session.config.ui_locator
    collector: Collector = session.collector

    # Create async wrapper functions that will be awaited
    async def _click(target_description: str):
        await click(cr, collector, target_description)

    async def _type(target_description: str, content: str):
        await type(cr, collector, target_description, content)

    async def _drag(source_description: str, target_description: str):
        await drag(cr, collector, source_description, target_description)

    async def _scroll(target_description: str | None, direction: str):
        await scroll(cr, collector, target_description, direction)

    async def _wait(duration: int):
        await wait(duration)

    async def _finished():
        await finished()

    async def _playwright_click(target_description: str):
        await playwright_click(target_description)

    # Remove triple backticks and optional 'python' tag
    try:
        _set_session(
            session
        )  # Bind this run to the session (page accessed dynamically)

        cleaned_code = re.sub(
            r"^```(?:python)?|```$", "", code.strip(), flags=re.MULTILINE
        ).strip()

        # Execute the code in an async context
        # We need to wrap the exec in an async function and await the calls
        local_vars: dict[str, Any] = {
            "click": _click,
            "type": _type,
            "drag": _drag,
            "scroll": _scroll,
            "wait": _wait,
            "finished": _finished,
            "playwright_click": _playwright_click,
        }

        # Parse the code and wrap each call in await
        tree = ast.parse(cleaned_code)

        # Execute each statement
        for node in tree.body:
            if isinstance(node, ast.Expr) and isinstance(node.value, ast.Call):
                # Get function name
                func_name = (
                    node.value.func.id
                    if isinstance(node.value.func, ast.Name)
                    else None
                )
                if func_name and func_name in local_vars:
                    # Extract arguments
                    args = []
                    for arg in node.value.args:
                        if isinstance(arg, ast.Constant):
                            args.append(arg.value)
                        else:
                            # Evaluate the argument
                            args.append(
                                eval(compile(ast.Expression(arg), "<string>", "eval"))
                            )

                    kwargs = {}
                    for kw in node.value.keywords:
                        if isinstance(kw.value, ast.Constant):
                            kwargs[kw.arg] = kw.value.value
                        else:
                            kwargs[kw.arg] = eval(
                                compile(ast.Expression(kw.value), "<string>", "eval")
                            )

                    # Call the async function
                    await local_vars[func_name](*args, **kwargs)
    finally:
        trace = deepcopy(_trace)
        _trace.clear()

    return trace



import base64
import logging
import time
from typing import Optional

from baml_py import Image
from baml_py.baml_py import BamlImagePy

from src.baml_client.async_client import b
from src.baml_client.types import Feedback
from src.cache import ActionCache, VerifyPostconditionCache
from src.config import Config

from .assertion_api import execute_assertion
from .assertion_api.session import Session

logger = logging.getLogger(__name__)


class BugReport(Exception):
    def __init__(self, message, screenshots=None, steps=None):
        super().__init__(message)
        self.screenshots = screenshots or []
        self.steps = steps or []


async def verify_precondition(
    session: Session, condition: str, action: str, config: Config
) -> None:
    logger.info(f"Condition: {condition}")
    client_registry = config.assertion_generation
    collector = session.collector
    screenshot_b64 = base64.b64encode(await session.page.screenshot(type="png")).decode(
        "utf-8"
    )
    screenshot = Image.from_base64("image/png", screenshot_b64)
    history = session.get_history()

    response = await b.GeneratePrecondition(
        screenshot,
        history,
        action,
        condition,
        feedback=[],
        baml_options={"client_registry": client_registry, "collector": collector},
    )
    passed, message = execute_assertion(response, session)
    if passed:
        logger.info("Precondition passed.")
        return
    else:
        logger.info(f"Precondition failed: {message}")
        raise BugReport(message)


async def execute_action(
    session: Session, action: str, config: Config, step_number: Optional[int] = None
) -> None:
    # Increment step counter and output step info for VSCode parsing
    if step_number is None:
        session.step_counter += 1
        step_number = session.step_counter

    # Record timestamp before current action.
    session.action_timestamps.append(time.time() - session.start_time)

    await session.update_mcp_progress(f"Reasoning next action {step_number}...")

    logger.info(f"Action: {action}")
    # Also output in a format that's easy for VSCode to parse
    logger.info(f"STEP_{step_number}: {action}")

    client_registry = config.action_proposer
    client_name = config.action_proposer_name
    collector = session.collector
    screenshot_b64 = base64.b64encode(await session.page.screenshot(type="png")).decode(
        "utf-8"
    )
    screenshot: BamlImagePy = Image.from_base64("image/png", screenshot_b64)

    # Check cache for proposed code
    cached = session.cache.get_action(step_number, action)
    if cached:
        code = cached.proposed_code
        logger.info(f"Cache hit for action at step {step_number}/{len(session.steps)}")
        logger.info(f"Proposed code (cached):\n{code}")
    else:
        logger.debug("Reasoning next action...")
        code = await b.ProposeActions(
            screenshot,
            action,
            baml_options={"client_registry": client_registry, "collector": collector},
        )
        logger.info(f"Proposed code:\n{code}")

    # Save to cache
    session.cache.set_action(
        step_number,
        ActionCache(screenshot=screenshot_b64, action=action, proposed_code=code),
    )

    try:
        if client_name == "UITARS":
            from .automators import uitars as uitars_automator

            trace = await uitars_automator.execute(code, session.page)
        elif client_name == "InternVL3":
            from .automators import pyautogui as pyautogui_automator

            trace = await pyautogui_automator.execute(code, session.page)
        else:
            from .automators import custom as custom_automator

            trace = await custom_automator.execute(code, session.page, session)

        session.trace.extend(trace)
        await session.update_mcp_progress(f"Action successful step {step_number}!")
        logger.info(f"STEP_{step_number}_PASSED")
        print(f"Action at step {step_number} executed successfully.")
    except Exception as e:
        logger.info(f"STEP_{step_number}_FAILED: {str(e)}")
        print(
            f"Action at step {step_number} failed, please try to be more specific about the element."
        )
        raise


async def verify_postcondition(
    session: Session,
    action: str,
    assertion: str,
    config: Config,
    step_number: Optional[int] = None,
) -> None:
    if step_number is None:
        step_number = session.step_counter

    # For VSCode & MCP progress
    logger.info(f"Expectation: {assertion}")
    logger.info(f"VERIFYING_STEP_{step_number}: {assertion}")
    await session.update_mcp_progress(f"Verifying post-condition step {step_number}...")

    client_registry = config.assertion_generation
    collector = session.collector
    max_tries = config.max_tries
    history = session.get_history()

    screenshot_after_b64 = session.history[-1].screenshot
    screenshot_after = Image.from_base64("image/png", screenshot_after_b64)

    # Check cache for postcondition
    cached = session.cache.get_verify_postcondition(step_number, assertion)
    if cached:
        postcondition = cached.proposed_code
        logger.info(f"Cache hit for verify_postcondition at step {step_number}")
        logger.info(f"Postcondition (cached): {postcondition}")
        logger.info("Executing assertions...")
        passed, message = execute_assertion(postcondition, session)

        # Save to cache
        if passed:
            logger.info("Postcondition passed.")
            logger.info(f"VERIFYING_STEP_{step_number}_PASSED")
            await session.update_mcp_progress(
                f"Post-condition verified successfully step {step_number}!"
            )
            print(f"Verification PASSED, {step_number}/{len(session.steps)}")
            session.cache.set_verify_postcondition(
                step_number,
                VerifyPostconditionCache.create(
                    assertion=assertion,
                    action=action,
                    proposed_code=postcondition,
                    history=history,
                    screenshot=screenshot_after_b64,
                ),
            )
            return
        else:
            logger.warning(f"Cached postcondition failed: {message}. Re-generating...")

    feedback: list[Feedback] = []
    message = "Post-condition verification failed after all retries"

    for _ in range(1, max_tries + 1):
        postcondition = await b.GeneratePostcondition(
            screenshot_after,
            history,
            action,
            assertion,
            feedback=feedback,
            baml_options={"client_registry": client_registry, "collector": collector},
        )
        logger.info(f"Postcondition: {postcondition}")
        logger.info("Executing assertions...")
        passed, message = execute_assertion(postcondition, session)
        if passed:
            logger.info("Postcondition passed.")
            logger.info(f"VERIFYING_STEP_{step_number}_PASSED")
            await session.update_mcp_progress(
                f"Post-condition verified successfully step {step_number}!"
            )

            # Save to cache
            session.cache.set_verify_postcondition(
                step_number,
                VerifyPostconditionCache.create(
                    assertion=assertion,
                    action=action,
                    proposed_code=postcondition,
                    history=history,
                    screenshot=screenshot_after_b64,
                ),
            )
            return
        else:
            logger.info(f"Postcondition failed: {message}")
            feedback_item = Feedback(response=postcondition, reason=message)
            feedback.append(feedback_item)

    logger.info(f"VERIFYING_STEP_{step_number}_FAILED: {message}")
    print(f"Verification at step {step_number} failed: {message}")
    session.failed_step = step_number
    raise BugReport(message)



from hashlib import sha1
from collections import defaultdict
from xml.etree.ElementTree import Element as XMLElement, indent, tostring

from .accessibility import AccessibilityTree


def _convert_node_to_xml(node: dict, parent: XMLElement = None) -> XMLElement:
    """
    Convert a node in the accessibility tree into an XML XMLElement.
    """
    role_value = node["role"]["value"]
    id = node.get("id", "")
    ignored = node.get("ignored", False)
    name_value = node.get("name", {}).get("value", "")
    properties = node.get("properties", [])
    children = node.get("nodes", [])

    if role_value == "StaticText":
        parent.text = name_value
    elif role_value == "none" or ignored:
        if children:
            for child in children:
                _convert_node_to_xml(child, parent)
    elif role_value == "generic" and not children:
        return None
    else:
        # Create the XML element for the node
        xml_element = XMLElement(role_value)

        if name_value and len(name_value) > 0:
            xml_element.set("name", name_value)

        # Assign a unique ID to the element
        xml_element.set("id", str(id))

        if properties:
            for property in properties:
                xml_element.set(
                    property["name"],
                    str(property.get("value", {}).get("value", "")),
                )

        # Add children recursively
        if children:
            for child in children:
                _convert_node_to_xml(child, xml_element)

        if parent is not None:
            parent.append(xml_element)

        return xml_element


def _get_texts(node: XMLElement) -> list[str]:
    """
    Get the text content of an element.
    """
    texts = set()
    if node.get("name"):
        texts.add(node.get("name"))
    if node.get("label"):
        texts.add(node.get("label"))
    if node.text:
        texts.add(node.text)

    return list(texts)


def _prune_redundant_name(node: XMLElement) -> list[str]:
    """
    Recursively traverses the tree, removes redundant name information from parent nodes,
    and returns a list of all content (names) in the current subtree.
    """
    # Remove name if it equals text
    if node.get("name") and node.text and node.get("name") == node.text:
        del node.attrib["name"]

    if not len(node):
        return _get_texts(node)

    # Recursively process children and gather all descendant content
    descendant_content = []
    for child in node:
        descendant_content.extend(_prune_redundant_name(child))

    # Sort by length, longest first, to handle overlapping substrings correctly
    descendant_content.sort(key=len, reverse=True)

    for content in descendant_content:
        if node.get("name"):
            node.set("name", node.get("name").replace(content, "").strip())
        if node.get("label"):
            node.set("label", node.get("label").replace(content, "").strip())
        if node.text:
            node.text = node.text.replace(content, "").strip()

    # The content of the current subtree is its own (potentially pruned) name
    # plus all the content from its descendants.
    current_subtree_content = descendant_content
    if node.get("name"):
        current_subtree_content.extend(_get_texts(node))

    return current_subtree_content


def _hash_element(elem: XMLElement, cache: dict[str, str]) -> str:
    """
    Hash an XMLElement and its subtree structure using the element's 'id'.
    A per-tree `cache` must be provided to avoid cross-tree contamination.
    """
    elem_id = elem.get("id")
    if not elem_id:
        raise ValueError("XMLElement is missing 'id' attribute")

    if elem_id in cache:
        return cache[elem_id]

    structure = elem.tag
    for child in elem:
        structure += _hash_element(child, cache)

    digest = sha1(structure.encode()).hexdigest()
    cache[elem_id] = digest
    return digest


def _group_similar_children(node: XMLElement, cache: dict[str, str]):
    """
    Groups structurally similar children under a <dynamic> node.
    """
    grouped = defaultdict(list)
    for child in list(node):  # Copy to avoid mutation during iteration
        try:
            child_hash = _hash_element(child, cache)
            grouped[child_hash].append(child)
        except ValueError:
            continue  # Skip children without 'id'

    for group in grouped.values():
        if len(group) >= 2:
            # Remove children from original node
            for child in group:
                node.remove(child)

            # Create a Dynamic wrapper and reattach the group
            dynamic_elem = XMLElement("dynamic")
            dynamic_elem.set("id", str(int(node.get("id", "0")) + 1000))
            dynamic_elem.set("summary", f"{len(group)} similar items")
            for child in group:
                dynamic_elem.append(child)

            node.append(dynamic_elem)

    # Recurse into children that are not Dynamic
    for child in node:
        if child.tag != "dynamic":
            _group_similar_children(child, cache)


def to_xml_tree(tree: AccessibilityTree) -> list[XMLElement]:
    """
    Abstracted tree for page reidentification.
    """
    if not isinstance(tree, AccessibilityTree):
        raise ValueError("Expected AccessibilityTree")

    if len(tree.root) != 1:
        raise ValueError(f"Expected 1 root node, got {len(tree.root)}")

    ((_, root_node),) = tree.root.items()
    element = _convert_node_to_xml(root_node)
    _prune_redundant_name(element)
    _group_similar_children(element, cache={})
    return element


def to_xml_string(tree: XMLElement) -> str:
    """
    String representation of abstracted tree for prompting.
    """
    if not isinstance(tree, XMLElement):
        raise ValueError("Expected xml.etree.XMLElementTree.XMLElement")

    indent(tree)
    return tostring(tree, encoding="unicode")



from typing import Any

from playwright.async_api import Page


class AccessibilityTree:
    """
    A model representing the accessibility tree of a Playwright `Page`.

    This tree is extracted using the Chrome DevTools Protocol (CDP)
    via `Accessibility.getFullAXTree`, and is reconstructed as a proper
    parent-child structure in memory.
    """

    def __init__(self):
        """
        Initialize an empty AccessibilityTree.
        Use the async `create` class method to properly initialize.
        """
        self.cdp_session = None
        self.root: dict[str, dict] = {}  # nodeId -> node (only roots)
        self.mapping: dict[int, Any] = {}  # internal node ID -> backendDOMNodeId

    @classmethod
    async def create(cls, page: Page) -> "AccessibilityTree":
        """
        Create and initialize an AccessibilityTree by retrieving and building the tree
        from the root document's accessibility structure.

        Args:
            page (Page): The Playwright page to extract the accessibility tree from.
            
        Returns:
            AccessibilityTree: A fully initialized accessibility tree.
        """
        instance = cls()
        instance.cdp_session = await page.context.new_cdp_session(page)
        nodes = await instance._fetch_accessibility_nodes()
        instance._build_tree(nodes)
        return instance

    async def _fetch_accessibility_nodes(self) -> list[dict]:
        """
        Fetch raw accessibility nodes from the CDP.

        Returns:
            List of accessibility node dictionaries.
        """
        response = await self.cdp_session.send("Accessibility.getFullAXTree")
        return response.get("nodes", [])

    def _build_tree(self, nodes: list[dict]):
        """
        Construct the in-memory tree from flat node list, identifying roots
        and building parent-child relationships.

        Args:
            nodes: List of node dictionaries from the accessibility snapshot.
        """
        node_lookup: dict[str, dict] = {node["nodeId"]: node for node in nodes}

        for internal_id, (node_id, node) in enumerate(node_lookup.items()):
            node["id"] = internal_id
            self.mapping[internal_id] = node.get("backendDOMNodeId", "")

            parent_id = node.get("parentId")

            if parent_id is None:
                # This is a root node (no parent in accessibility tree)
                self.root[node_id] = node
            else:
                # Attach this node to its parent
                parent = node_lookup[parent_id]
                parent.setdefault("nodes", []).append(node)

                # Clean up linkage fields from the child
                node.pop("childIds", None)
                node.pop("parentId", None)



import argparse
import asyncio
import json
import logging
import sys
from pathlib import Path

from playwright.async_api import async_playwright

from src.run import run_test_from_file

logger = logging.getLogger(__name__)
VIEWPORT_WIDTH = 1920
VIEWPORT_HEIGHT = 1080


def main():
    """Main CLI entry point"""
    parser = argparse.ArgumentParser(
        description="WebTestPilot CLI - Run automated web tests from VS Code"
    )

    parser.add_argument("test_file", type=str, help="Path to the test JSON file")

    parser.add_argument(
        "--config",
        type=str,
        default="config.yaml",
        help="Path to config.yaml file (default: config.yaml)",
    )

    parser.add_argument(
        "--cdp-endpoint",
        type=str,
        default="http://localhost:9222",
        help="CDP endpoint URL (default: http://localhost:9222)",
    )

    parser.add_argument(
        "--target-id",
        type=str,
        help="id of cdp session",
    )

    parser.add_argument(
        "--no-assertions",
        action="store_true",
        help="Disable assertion verification (only execute actions)",
    )

    parser.add_argument(
        "--fixture-file-path",
        type=str,
        default="",
        help="Path to fixture file",
    )

    parser.add_argument(
        "--environment-file-path",
        type=str,
        default="",
        help="Path to environment file",
    )

    parser.add_argument(
        "--cache-path",
        type=str,
        default="",
        help="Path to cache file (e.g., .webtestpilot/.cache/test_id.json)",
    )

    args = parser.parse_args()
    cache_path = Path(args.cache_path) if args.cache_path else None

    # Run the test
    async def run():
        async with async_playwright() as playwright:
            browser = await playwright.chromium.connect_over_cdp(args.cdp_endpoint)
            return await run_test_from_file(
                browser=browser,
                test_file_path=args.test_file,
                config_path=args.config,
                target_id=args.target_id,
                fixture_file_path=args.fixture_file_path,
                environment_file_path=args.environment_file_path,
                enable_assertions=not args.no_assertions,
                cache_path=cache_path,
            )
    
    result = asyncio.run(run())

    print(json.dumps(result, indent=2))

    # Exit with appropriate code
    logger.debug(f"Exiting with code: {0 if result['success'] else 1}")
    sys.exit(0 if result["success"] else 1)


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------------
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
# ----------------------------------------------------------------------------

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code using: baml-cli generate
# baml-cli is available with the baml package.

from . import types
from . import stream_types


type_map = {

    "types.Feedback": types.Feedback,
    "stream_types.Feedback": stream_types.Feedback,

    "types.History": types.History,
    "stream_types.History": stream_types.History,

    "types.Output": types.Output,
    "stream_types.Output": stream_types.Output,

    "types.PageAbstract": types.PageAbstract,
    "stream_types.PageAbstract": stream_types.PageAbstract,

    "types.Step": types.Step,
    "stream_types.Step": stream_types.Step,

    "types.TestCase": types.TestCase,
    "stream_types.TestCase": stream_types.TestCase,


}


import json
import logging
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Optional

from src.baml_client.types import History

logger = logging.getLogger(__name__)


@dataclass
class PageContent:
    name: str
    description: str
    layout: str


@dataclass
class CaptureStateCache:
    screenshot: str
    page_content: PageContent
    prev_action: Optional[str] = None  # Track what action led to this state


@dataclass
class ActionCache:
    screenshot: str
    action: str
    proposed_code: str


@dataclass
class HistoryItem:
    page_id: str
    layout: Optional[str]
    description: Optional[str]
    prev_action: Optional[str]


@dataclass
class VerifyPostconditionCache:
    assertion: str
    action: str
    proposed_code: str
    history: list[HistoryItem]
    screenshot: str

    @classmethod
    def create(
        cls,
        assertion: str,
        action: str,
        proposed_code: str,
        history: list[History],
        screenshot: str,
    ) -> "VerifyPostconditionCache":
        history_items = [
            HistoryItem(
                page_id=h.page_id,
                layout=h.layout,
                description=h.description,
                prev_action=h.prev_action,
            )
            for h in history
        ]
        return cls(
            assertion=assertion,
            action=action,
            proposed_code=proposed_code,
            history=history_items,
            screenshot=screenshot,
        )


class Cache:
    """Simple cache class for storing and retrieving test execution results."""

    def __init__(self, cache_path: Optional[Path] = None):
        self.cache_path = cache_path
        self.data: dict[str, dict] = {}

        # Load existing cache if path provided
        if cache_path and cache_path.exists():
            try:
                with open(cache_path, "r", encoding="utf-8") as f:
                    self.data = json.load(f)
                logger.info(f"Loaded cache from: {cache_path}")
            except Exception as e:
                logger.warning(f"Failed to load cache: {e}")
                self.data = {}

    def get_capture_state(
        self, step_number: int, prev_action: Optional[str] = None
    ) -> Optional[CaptureStateCache]:
        """Get capture_state cache for a step, optionally validating prev_action matches.

        Args:
            step_number: The step number to get cache for.
            prev_action: The previous action to validate against cached prev_action.

        Returns:
            CaptureStateCache if cache exists and prev_action matches (if provided), None otherwise.
        """
        step_data = self.data.get(f"step_{step_number}", {})
        if "capture_state" not in step_data:
            return None

        cs = step_data["capture_state"]
        cached = CaptureStateCache(
            screenshot=cs["screenshot"],
            page_content=PageContent(**cs["page_content"]),
            prev_action=cs.get(
                "prev_action"
            ),  # Handle old cache format without prev_action
        )

        # Validate prev_action matches if provided
        if prev_action is not None and cached.prev_action != prev_action:
            logger.debug(
                f"prev_action mismatch at step {step_number}: cached='{cached.prev_action}' vs current='{prev_action}'"
            )
            return None

        return cached

    def set_capture_state(self, step_number: int, cache: CaptureStateCache) -> None:
        """Set capture_state cache for a step."""
        step_key = f"step_{step_number}"
        if step_key not in self.data:
            self.data[step_key] = {}
        self.data[step_key]["capture_state"] = asdict(cache)

    def get_action(self, step_number: int, action: str) -> Optional[ActionCache]:
        """Get action cache for a step, validating action matches.

        Args:
            step_number: The step number to get cache for.
            action: The action string to validate against cached action.

        Returns:
            ActionCache if cache exists and action matches, None otherwise.
        """
        step_data = self.data.get(f"step_{step_number}", {})
        if "action" not in step_data:
            return None

        cached = ActionCache(**step_data["action"])
        # Validate action matches
        if cached.action != action:
            logger.debug(
                f"Action mismatch at step {step_number}: cached='{cached.action}' vs current='{action}'"
            )
            return None

        return cached

    def set_action(self, step_number: int, cache: ActionCache) -> None:
        """Set action cache for a step."""
        step_key = f"step_{step_number}"
        if step_key not in self.data:
            self.data[step_key] = {}
        self.data[step_key]["action"] = asdict(cache)

    def get_verify_postcondition(
        self, step_number: int, assertion: str
    ) -> Optional[VerifyPostconditionCache]:
        """Get verify_postcondition cache for a step, validating assertion matches.

        Args:
            step_number: The step number to get cache for.
            assertion: The assertion/expectation string to validate against cached assertion.

        Returns:
            VerifyPostconditionCache if cache exists and assertion matches, None otherwise.
        """
        step_data = self.data.get(f"step_{step_number}", {})
        if "verify_postcondition" not in step_data:
            return None

        vpc = step_data["verify_postcondition"]
        cached = VerifyPostconditionCache(
            assertion=vpc["assertion"],
            action=vpc["action"],
            proposed_code=vpc["proposed_code"],
            history=[HistoryItem(**h) for h in vpc["history"]],
            screenshot=vpc["screenshot"],
        )
        # Validate assertion matches
        if cached.assertion != assertion:
            logger.debug(
                f"Assertion mismatch at step {step_number}: cached='{cached.assertion}' vs current='{assertion}'"
            )
            return None

        return cached

    def set_verify_postcondition(
        self, step_number: int, cache: VerifyPostconditionCache
    ) -> None:
        """Set verify_postcondition cache for a step."""
        step_key = f"step_{step_number}"
        if step_key not in self.data:
            self.data[step_key] = {}
        self.data[step_key]["verify_postcondition"] = asdict(cache)

    def save(self) -> None:
        """Save cache to file."""
        if not self.cache_path:
            logger.warning("No cache path specified, skipping save")
            return

        try:
            self.cache_path.parent.mkdir(parents=True, exist_ok=True)
            with open(self.cache_path, "w", encoding="utf-8") as f:
                json.dump(self.data, f, indent=2, ensure_ascii=False)
            logger.info(f"Saved cache to: {self.cache_path}")
        except Exception as e:
            logger.error(f"Failed to save cache: {e}")

    def save_at_step(self, step_number: int) -> None:
        """Save cache to file."""
        if not self.cache_path:
            logger.warning("No cache path specified, skipping save")
            return

        try:
            self.cache_path.parent.mkdir(parents=True, exist_ok=True)
            # change name from cache.json to cache_step_{step_number}.json

            with open(
                self.cache_path.parent / f"cache_step_{step_number}.json",
                "w",
                encoding="utf-8",
            ) as f:
                json.dump(self.data, f, indent=2, ensure_ascii=False)
            logger.info(f"Saved cache to: {self.cache_path}")
        except Exception as e:
            logger.error(f"Failed to save cache: {e}")



import logging
import re
import sys
import pprint
from typing import Any, Callable, Union, Literal, Optional, List, Set, Dict
from enum import Enum
from uuid import UUID
from datetime import datetime, date

from pydantic import BaseModel, Field, EmailStr, HttpUrl, constr, conint, confloat

from .session import Session
from .state import State
from .element import Element


PRIMITIVE_TYPES = (int, float, str, bool, type(None))
logger = logging.getLogger(__name__)

def _run_assertion_with_trace(
    assertion_func: Callable[[Session], Any], session: Session
) -> tuple[bool, str]:
    """
    Executes a given assertion function and captures variable traces
    (only primitives, dicts, lists, or BaseModel instances).
    """
    captured_vars = []

    def filter_locals(locals_dict):
        filtered = {}
        for k, v in locals_dict.items():
            if isinstance(v, PRIMITIVE_TYPES):
                filtered[k] = v
            elif isinstance(v, (dict, list)):
                filtered[k] = v
            elif isinstance(v, BaseModel):
                filtered[k] = v.dict()
        return filtered

    def tracer(frame, event, arg):
        if event == "line" and frame.f_code.co_name == assertion_func.__name__:
            # Filter locals to only include allowed types
            captured_vars.append((frame.f_lineno, filter_locals(frame.f_locals.copy())))
        return tracer

    try:
        sys.settrace(tracer)
        result = assertion_func(session)
        sys.settrace(None)

        if result is True or result is None:
            return True, "Success."

        if isinstance(result, str) and result.strip():
            return False, result

        # return (
        #     False,
        #     f"Assertion failed without message.\nVariable trace:\n{pprint.pformat(captured_vars)}",
        # )
        return (
            False,
            "Assertion failed without message.",
        )

    except AssertionError as ae:
        sys.settrace(None)
        msg = str(ae) if str(ae).strip() else "AssertionError raised without message."
        # return False, f"{msg}\nVariable trace:\n{pprint.pformat(captured_vars)}"
        return False, f"{msg}"

    except Exception:
        sys.settrace(None)
        raise


def execute_assertion(response: str, session: Session) -> tuple[bool, str]:
    """
    Executes assertion code blocks extracted from response.

    Looks for code blocks formatted as ```python ... ``` and expects
    either `precondition(session)` or `postcondition(session)`.

    Returns (success, message). On failure, includes variable trace for debugging.
    """
    # Extract python code blocks
    pattern = r"```python\s+([\s\S]*?)```"
    code_blocks = re.findall(pattern, response, re.MULTILINE)
    code = "\n\n".join(code_blocks)

    # Prepare safe globals for exec
    allowed_globals = {
        "__builtins__": __builtins__,
        "Session": Session,
        "State": State,
        "Element": Element,
        # Pydantic
        "BaseModel": BaseModel,
        "Field": Field,
        "EmailStr": EmailStr,
        "HttpUrl": HttpUrl,
        "constr": constr,
        "conint": conint,
        "confloat": confloat,
        # Typing
        "Any": Any,
        "Union": Union,
        "Literal": Literal,
        "Optional": Optional,
        "List": List,
        "Set": Set,
        "Dict": Dict,
        # Enum
        "Enum": Enum,
        # Common value types
        "UUID": UUID,
        "datetime": datetime,
        "date": date,
    }

    local_vars: dict = {}

    # Execute the dynamic code
    exec(code, allowed_globals, local_vars)

    # Find the assertion function
    assertion_func = local_vars.get("precondition") or local_vars.get("postcondition")
    if assertion_func is None or not callable(assertion_func):
        return (
            False,
            "No callable 'precondition' or 'postcondition' function found in generated code.",
        )

    # Run the assertion with variable tracing
    return _run_assertion_with_trace(assertion_func, session)



# ----------------------------------------------------------------------------
#
#  Welcome to Baml! To use this generated code, please run the following:
#
#  $ pip install baml
#
# ----------------------------------------------------------------------------

# This file was generated by BAML: please do not edit it. Instead, edit the
# BAML files and re-generate this code using: baml-cli generate
# baml-cli is available with the baml package.

import typing
from baml_py import type_builder
from baml_py import baml_py
# These are exports, not used here, hence the linter is disabled
from baml_py.baml_py import FieldType, EnumValueBuilder, EnumBuilder, ClassBuilder # noqa: F401 # pylint: disable=unused-import
from .globals import DO_NOT_USE_DIRECTLY_UNLESS_YOU_KNOW_WHAT_YOURE_DOING_RUNTIME

class TypeBuilder(type_builder.TypeBuilder):
    def __init__(self):
        super().__init__(classes=set(
          ["Feedback","History","Output","PageAbstract","Step","TestCase",]
        ), enums=set(
          []
        ), runtime=DO_NOT_USE_DIRECTLY_UNLESS_YOU_KNOW_WHAT_YOURE_DOING_RUNTIME)

    # #########################################################################
    # Generated enums 0
    # #########################################################################


    # #########################################################################
    # Generated classes 6
    # #########################################################################

    @property
    def Feedback(self) -> "FeedbackViewer":
        return FeedbackViewer(self)

    @property
    def History(self) -> "HistoryViewer":
        return HistoryViewer(self)

    @property
    def Output(self) -> "OutputBuilder":
        return OutputBuilder(self)

    @property
    def PageAbstract(self) -> "PageAbstractViewer":
        return PageAbstractViewer(self)

    @property
    def Step(self) -> "StepViewer":
        return StepViewer(self)

    @property
    def TestCase(self) -> "TestCaseViewer":
        return TestCaseViewer(self)



# #########################################################################
# Generated enums 0
# #########################################################################


# #########################################################################
# Generated classes 6
# #########################################################################

class FeedbackAst:
    def __init__(self, tb: type_builder.TypeBuilder):
        _tb = tb._tb # type: ignore (we know how to use this private attribute)
        self._bldr = _tb.class_("Feedback")
        self._properties: typing.Set[str] = set([  "response",  "reason",  ])
        self._props = FeedbackProperties(self._bldr, self._properties)

    def type(self) -> baml_py.FieldType:
        return self._bldr.field()

    @property
    def props(self) -> "FeedbackProperties":
        return self._props


class FeedbackViewer(FeedbackAst):
    def __init__(self, tb: type_builder.TypeBuilder):
        super().__init__(tb)

    
    def list_properties(self) -> typing.List[typing.Tuple[str, type_builder.ClassPropertyViewer]]:
        return [(name, type_builder.ClassPropertyViewer(self._bldr.property(name))) for name in self._properties]
    


class FeedbackProperties:
    def __init__(self, bldr: baml_py.ClassBuilder, properties: typing.Set[str]):
        self.__bldr = bldr
        self.__properties = properties # type: ignore (we know how to use this private attribute) # noqa: F821

    
    
    @property
    def response(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("response"))
    
    @property
    def reason(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("reason"))
    
    


class HistoryAst:
    def __init__(self, tb: type_builder.TypeBuilder):
        _tb = tb._tb # type: ignore (we know how to use this private attribute)
        self._bldr = _tb.class_("History")
        self._properties: typing.Set[str] = set([  "page_id",  "layout",  "description",  "prev_action",  ])
        self._props = HistoryProperties(self._bldr, self._properties)

    def type(self) -> baml_py.FieldType:
        return self._bldr.field()

    @property
    def props(self) -> "HistoryProperties":
        return self._props


class HistoryViewer(HistoryAst):
    def __init__(self, tb: type_builder.TypeBuilder):
        super().__init__(tb)

    
    def list_properties(self) -> typing.List[typing.Tuple[str, type_builder.ClassPropertyViewer]]:
        return [(name, type_builder.ClassPropertyViewer(self._bldr.property(name))) for name in self._properties]
    


class HistoryProperties:
    def __init__(self, bldr: baml_py.ClassBuilder, properties: typing.Set[str]):
        self.__bldr = bldr
        self.__properties = properties # type: ignore (we know how to use this private attribute) # noqa: F821

    
    
    @property
    def page_id(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("page_id"))
    
    @property
    def layout(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("layout"))
    
    @property
    def description(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("description"))
    
    @property
    def prev_action(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("prev_action"))
    
    


class OutputAst:
    def __init__(self, tb: type_builder.TypeBuilder):
        _tb = tb._tb # type: ignore (we know how to use this private attribute)
        self._bldr = _tb.class_("Output")
        self._properties: typing.Set[str] = set([  ])
        self._props = OutputProperties(self._bldr, self._properties)

    def type(self) -> baml_py.FieldType:
        return self._bldr.field()

    @property
    def props(self) -> "OutputProperties":
        return self._props


class OutputBuilder(OutputAst):
    def __init__(self, tb: type_builder.TypeBuilder):
        super().__init__(tb)

    
    def add_property(self, name: str, type: baml_py.FieldType) -> baml_py.ClassPropertyBuilder:
        if name in self._properties:
            raise ValueError(f"Property {name} already exists.")
        return self._bldr.property(name).type(type)

    def list_properties(self) -> typing.List[typing.Tuple[str, baml_py.ClassPropertyBuilder]]:
        return self._bldr.list_properties()

    def remove_property(self, name: str) -> None:
        self._bldr.remove_property(name)

    def reset(self) -> None:
        self._bldr.reset()

    


class OutputProperties:
    def __init__(self, bldr: baml_py.ClassBuilder, properties: typing.Set[str]):
        self.__bldr = bldr
        self.__properties = properties # type: ignore (we know how to use this private attribute) # noqa: F821

    
    def __getattr__(self, name: str) -> baml_py.ClassPropertyBuilder:
        if name not in self.__properties:
            raise AttributeError(f"Property {name} not found.")
        return self.__bldr.property(name)

    
    


class PageAbstractAst:
    def __init__(self, tb: type_builder.TypeBuilder):
        _tb = tb._tb # type: ignore (we know how to use this private attribute)
        self._bldr = _tb.class_("PageAbstract")
        self._properties: typing.Set[str] = set([  "name",  "description",  "layout",  ])
        self._props = PageAbstractProperties(self._bldr, self._properties)

    def type(self) -> baml_py.FieldType:
        return self._bldr.field()

    @property
    def props(self) -> "PageAbstractProperties":
        return self._props


class PageAbstractViewer(PageAbstractAst):
    def __init__(self, tb: type_builder.TypeBuilder):
        super().__init__(tb)

    
    def list_properties(self) -> typing.List[typing.Tuple[str, type_builder.ClassPropertyViewer]]:
        return [(name, type_builder.ClassPropertyViewer(self._bldr.property(name))) for name in self._properties]
    


class PageAbstractProperties:
    def __init__(self, bldr: baml_py.ClassBuilder, properties: typing.Set[str]):
        self.__bldr = bldr
        self.__properties = properties # type: ignore (we know how to use this private attribute) # noqa: F821

    
    
    @property
    def name(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("name"))
    
    @property
    def description(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("description"))
    
    @property
    def layout(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("layout"))
    
    


class StepAst:
    def __init__(self, tb: type_builder.TypeBuilder):
        _tb = tb._tb # type: ignore (we know how to use this private attribute)
        self._bldr = _tb.class_("Step")
        self._properties: typing.Set[str] = set([  "condition",  "action",  "expectation",  ])
        self._props = StepProperties(self._bldr, self._properties)

    def type(self) -> baml_py.FieldType:
        return self._bldr.field()

    @property
    def props(self) -> "StepProperties":
        return self._props


class StepViewer(StepAst):
    def __init__(self, tb: type_builder.TypeBuilder):
        super().__init__(tb)

    
    def list_properties(self) -> typing.List[typing.Tuple[str, type_builder.ClassPropertyViewer]]:
        return [(name, type_builder.ClassPropertyViewer(self._bldr.property(name))) for name in self._properties]
    


class StepProperties:
    def __init__(self, bldr: baml_py.ClassBuilder, properties: typing.Set[str]):
        self.__bldr = bldr
        self.__properties = properties # type: ignore (we know how to use this private attribute) # noqa: F821

    
    
    @property
    def condition(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("condition"))
    
    @property
    def action(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("action"))
    
    @property
    def expectation(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("expectation"))
    
    


class TestCaseAst:
    def __init__(self, tb: type_builder.TypeBuilder):
        _tb = tb._tb # type: ignore (we know how to use this private attribute)
        self._bldr = _tb.class_("TestCase")
        self._properties: typing.Set[str] = set([  "name",  "steps",  ])
        self._props = TestCaseProperties(self._bldr, self._properties)

    def type(self) -> baml_py.FieldType:
        return self._bldr.field()

    @property
    def props(self) -> "TestCaseProperties":
        return self._props


class TestCaseViewer(TestCaseAst):
    def __init__(self, tb: type_builder.TypeBuilder):
        super().__init__(tb)

    
    def list_properties(self) -> typing.List[typing.Tuple[str, type_builder.ClassPropertyViewer]]:
        return [(name, type_builder.ClassPropertyViewer(self._bldr.property(name))) for name in self._properties]
    


class TestCaseProperties:
    def __init__(self, bldr: baml_py.ClassBuilder, properties: typing.Set[str]):
        self.__bldr = bldr
        self.__properties = properties # type: ignore (we know how to use this private attribute) # noqa: F821

    
    
    @property
    def name(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("name"))
    
    @property
    def steps(self) -> type_builder.ClassPropertyViewer:
        return type_builder.ClassPropertyViewer(self.__bldr.property("steps"))
    
    




import logging
import logging.config
from dataclasses import dataclass
from pathlib import Path
from typing import Any

import yaml
from baml_py import ClientRegistry
from dotenv import load_dotenv

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Config:
    parser: ClientRegistry
    assertion_generation: ClientRegistry
    assertion_api: ClientRegistry
    action_proposer: ClientRegistry
    action_proposer_name: str
    ui_locator: ClientRegistry
    page_reidentification: ClientRegistry

    infer_missing_steps: bool
    max_tries: int

    @staticmethod
    def load(path: Path | str) -> "Config":
        # Load environment variables
        load_dotenv()

        # Load YAML config
        yaml_path = Path(path)
        with yaml_path.open("r") as f:
            yaml_data: dict[str, Any] = yaml.safe_load(f) or {}

        # LLM client configurations
        parser = ClientRegistry()
        parser.set_primary(yaml_data["parser"]["llm_client"])
        infer_missing_steps = yaml_data["parser"]["infer_missing_steps"]

        executor_clients = yaml_data["executor"]["llm_clients"]
        assertion_generation = ClientRegistry()
        assertion_generation.set_primary(executor_clients["assertion_generation"])

        assertion_api = ClientRegistry()
        assertion_api.set_primary(executor_clients["assertion_api"])
        action_proposer_name = executor_clients["action_proposer"]

        action_proposer = ClientRegistry()
        action_proposer.set_primary(action_proposer_name)

        ui_locator = ClientRegistry()
        ui_locator.set_primary(executor_clients["ui_locator"])

        page_reidentification = ClientRegistry()
        page_reidentification.set_primary(executor_clients["page_reidentification"])

        max_tries = yaml_data["executor"]["max_tries"]

        logging.config.dictConfig(yaml_data.get("logging", {}))

        return Config(
            parser=parser,
            assertion_generation=assertion_generation,
            assertion_api=assertion_api,
            action_proposer_name=action_proposer_name,
            action_proposer=action_proposer,
            ui_locator=ui_locator,
            page_reidentification=page_reidentification,
            infer_missing_steps=infer_missing_steps,
            max_tries=max_tries,
        )



from src.baml_client.sync_client import b
from src.baml_client.types import Step, TestCase
from src.config import Config


class MissingStepsError(Exception):
    def __init__(self, steps):
        super().__init__(
            f"Missing steps, please update test description to address <unknown> parts. Steps: {steps}"
        )
        self.steps = steps


def parse(description: str, config: Config) -> TestCase:
    """
    Parse a natural language test description into a structured test case.

    Args:
        description (str):
            A plain-text description of a test scenario, written in natural language.

    Returns:
        TestCase:
            A structured test case object containing:
            - scenario: A list of steps.
            - Each step has three components:
                * condition: Preconditions or setup.
                * action: The event or interaction being performed.
                * expectation: The expected outcome or assertion.
    """

    def contains_unknown(steps: list[Step]):
        return any(
            "<unknown>" in part
            for step in steps
            for part in (step.condition, step.action, step.expectation)
        )

    client_registry = config.parser
    test_case = b.ExtractTestCase(description, {"client_registry": client_registry})

    if contains_unknown(test_case.steps):
        if config.infer_missing_steps:
            return b.ImplicitGeneration(test_case, {"client_registry": client_registry})
        raise MissingStepsError(test_case.steps)

    return test_case
